//--------------------
// STYLE
//--------------------

var textcolour = RGB(255, 255, 255);
var backcolour = RGB(0, 0, 0);
var activecolour = RGB(128, 0, 255);
var checkedcolour = RGB(104, 104, 104);
var sepcolour = RGB(127, 127, 127);
var edittextcolour = RGB(0, 0, 0);
var editbackcolour = RGB(255, 255, 255);
var editsepcolour = RGB(127, 127, 127);
var uncheckedcolour = RGB(64, 64, 64);
var errorcolour = RGB(255, 0, 0);
var selectcolour = RGB(209, 163, 255);

var marginx = 5;
var marginy = 5;

var displaysep = ", ";
var ratingon = String.fromCharCode(0x2605);
var ratingoff = String.fromCharCode(0x2606);

//--------------------
// FORM ELEMENTS
//--------------------

//--------------------
// General comments:
//
// - If 'x', 'y', 'w', or 'h' depends on the size of the screen, this value should be placed between brackets, where 'ww' could be used for window.Width and 'wh' for window.Height.
//   For example: FormIButton("ww - 20", 0, 20, 20, skip, imgscr, func);
//   In case of a pop-up also 'px', 'py', 'pw' and 'ph' could be used, refering to the size and position of the pop-up
//
// - One can jump from element to element using shift-tab/tab, left/right and up/down keys, although if skip is set true the element will be skipped.
//
// - 'font' should be of type IGdiFont
//
// - 'func' should always be a function that if called without argument returns the value to be displayed and that will be called if this value is changed by the form element.
//
// - For single line text elements with a fixed width a tool tip shows the value on mouse over if clipping occurs.
//
// - These functions should be called in the corresponding callback functions:
//
//		forms_on_size();
//		forms_on_paint(gr);
//		forms_on_mouse_move(x, y);
//		forms_on_mouse_lbtn_down(x, y)
//		forms_on_mouse_lbtn_up(x, y);
//		forms_on_mouse_rbtn_down(x, y);
//		forms_on_mouse_rbtn_up(x, y);
//		forms_on_mouse_wheel(step);
//		forms_on_key_down(vkey);
//		forms_on_char(code);
//		forms_on_focus(is_focused);
//		forms_on_timer(id);

FormsAndButtons = new FAndB;

function FAndB () {
	//--------------------
	// function FormsAndButtons.AddIButton(x, y, w, h, skip, imgsrc[, func, rbfunc])
	//--------------------
	// Draws a button using images.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//	ENTER: switch button
	//
	// func: boolean function(boolean);
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);
	// imgsrc: array {inactive: <<path to image>>, active: <<path to image>>, inactiveon: <<path to image>>, activeon: <<path to image>>}

	this.AddIButton = function (x, y, w, h, skip, imgsrc, func, rbfunc) {this.list.push(new FormIButton(x, y, w, h, skip, imgsrc, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddTButton(x, y, w, h, skip, text, font[, func, rbfunc])
	//--------------------
	// Draws a button using text.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//	ENTER: switch button
	//
	// func: boolean function(boolean);
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);
	// if w is set 0 it will be based on marginx, if h is set 0 it will be based on marginy

	this.AddTButton = function (x, y, w, h, skip, text, font, func, rbfunc) {this.list.push(new FormTButton(x, y, w, h, skip, text, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddTButtonAlt(x, y, w, h, skip, text, font[, func, rbfunc])
	//--------------------
	// Draws a button which is nothing more than a line of text which can be set or not set.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//	ENTER: switch button
	//
	// func: boolean function(boolean);
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);

	this.AddTButtonAlt = function (x, y, w, h, skip, text, font, func, rbfunc) {this.list.push(new FormTButtonAlt(x, y, w, h, skip, text, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddTextBox(x, y, w, h, skip, font[, func, rbfunc])
	//--------------------
	// Draws a text box for editing multiple lines of text.
	//--------------------
	// Using keyboard (not in edit mode):
	//	SHIFT-TAB, LEFT: jump to previous element
	//	TAB, RIGHT: jump to next element
	//	ENTER:	enter edit mode
	//	PGUP: scroll to start
	//	PGDN: scroll to end
	//	UP:	if possible scroll up, otherwise jump to previous element
	//	DOWN: if possible scroll down, otherwise jump to next element
	//	
	// Using keyboard (edit mode):
	//	BKSP, ENTER, END, HOME, LEFT, UP, RIGHT, DOWN, INS, DEL, CTRL-A, CTRL-C, CTRL-V, CTRL-X
	//		as expected in a text editor (hold SHIFT to select)
	//	SHIFT-TAB: exit edit mode, send value if changed and jump to previous element
	//	TAB: exit edit mode, send value if changed and jump to next element
	//	ESC: exit edit mode and send value if changed
	//	PGUP: cursor to start of text
	//	PGDN: cursor to end of text
	//	CTRL-Z:	undo change
	//
	// func: string function(string);
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);

	this.AddTextBox = function (x, y, w, h, skip, font, func, rbfunc) {this.list.push(new FormTextBox(x, y, w, h, skip, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddText(x, y, w, h, skip, font[, func, rbfunc])
	//--------------------
	// Draws multiple lines of text.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, LEFT: jump to previous element
	//	TAB, RIGHT: jump to next element
	//	PGUP: scroll to start
	//	PGDN: scroll to end
	//	UP: if possible scroll up, otherwise jump to previous element
	//	DOWN: if possible scroll down, otherwise jump to next element
	//	
	// func: string function(string);
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);

	this.AddText = function (x, y, w, h, skip, font, func, rbfunc) {this.list.push(new FormText(x, y, w, h, skip, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddTextInput(x, y, w, skip, font[, func, rbfunc])
	//--------------------
	// Draws single line text input.
	//--------------------
	// Using keyboard (not in edit mode):
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//	ENTER: enter edit mode
	//	
	// Using keyboard (edit mode):
	//	BKSP, END, HOME, LEFT, RIGHT, INS, DEL, CTRL-A, CTRL-C, CTRL-V, CTRL-X
	//		as expected in a text editor (hold SHIFT to select)
	//	SHIFT-TAB: exit edit mode, send value if changed and jump to previous element
	//	TAB: exit edit mode, send value if changed and jump to next element
	//	ENTER:	exit edit mode and send value if changed
	//	ESC: exit edit mode and send value if changed
	//	CTRL-Z: undo change
	//
	// func: string function(string);
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);

	this.AddTextInput = function (x, y, w, skip, font, func, rbfunc) {this.list.push(new FormTextInput(x, y, w, skip, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddTextLine(x, y, w, skip, font[, func, rbfunc])
	//--------------------
	// Draws single line of text.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//
	// func: string function();
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);

	this.AddTextLine = function (x, y, w, skip, font, func, rbfunc) {this.list.push(new FormTextLine(x, y, w, skip, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddMVTextInput(x, y, w, skip, font[, func, rbfunc])
	//--------------------
	// Draws single line text input to edit an array of values.
	//--------------------
	// Using keyboard (not in edit mode):
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//	ENTER: enter edit mode
	//	
	// Using keyboard (edit mode):
	//	BKSP, END, HOME, LEFT, RIGHT, INS, DEL, CTRL-A, CTRL-C, CTRL-V, CTRL-X
	//		as expected in a text editor (hold SHIFT to select)
	//	SHIFT-TAB: exit edit mode, send value if changed and jump to previous element
	//	TAB: exit edit mode, send value if changed and jump to next element
	//	ENTER:	exit edit mode and send value if changed
	//	ESC: exit edit mode and send value if changed
	//	CTRL-Z: undo change
	//
	// func: array function(array);
	// rbfunc: function(x, y, valuenr, hoverx, hovery, hoverw, hoverh);

	this.AddMVTextInput = function (x, y, w, skip, font, func, rbfunc) {this.list.push(new FormMVTextInput(x, y, w, skip, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddMVSTextInput(x, y, w, skip, font[, func, rbfunc])
	//--------------------
	// Draws single line text input to edit an array of values in which the odd ones are the main values and the even ones the separators.
	//--------------------
	// Using keyboard (not in edit mode):
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//	ENTER: enter edit mode
	//	
	// Using keyboard (edit mode):
	//	BKSP, END, HOME, LEFT, RIGHT, INS, DEL, CTRL-A, CTRL-C, CTRL-V, CTRL-X
	//		as expected in a text editor (hold SHIFT to select)
	//	SHIFT-TAB: exit edit mode, send value if changed and jump to previous element
	//	TAB: exit edit mode, send value if changed and jump to next element
	//	ENTER:	exit edit mode and send value if changed
	//	ESC: exit edit mode and send value if changed
	//	CTRL-Z: undo change
	//
	// func: array function(array); where this array contains the values as odd elements and the separators as even elements
	// rbfunc: function(x, y, valuenr, hoverx, hovery, hoverw, hoverh);

	this.AddMVSTextInput = function (x, y, w, skip, font, func, rbfunc) {this.list.push(new FormMVSTextInput(x, y, w, skip, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddMVTextLine(x, y, w, skip, font[, func, rbfunc])
	//--------------------
	// Draws single line showing the content of an array of values.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, LEFT, UP: jump to previous element
	//	TAB, RIGHT, DOWN: jump to next element
	//
	// func: array function();	where this array contains the values as odd elements and the separators as even elements
	// rbfunc: function(x, y, valuenr, hoverx, hovery, hoverw, hoverh);

	this.AddMVTextLine = function (x, y, w, skip, font, func, rbfunc) {this.list.push(new FormMVTextLine(x, y, w, skip, font, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddKeyWords(x, y, w, skip, font[, listfunc, func, rbfunc])
	//--------------------
	// Draws single line showing the content of an array of key words and add new ones from a drown-down list.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, UP: jump to previous element
	//	TAB, DOWN: jump to next element
	//	ENTER: execute selected button
	//	END: jump to last value
	//	HOME: jump to first value
	//	LEFT: if possible jump to previous button, otherwise jump to previous element
	//	CTRL-LEFT: move selected value to the left
	//	RIGHT:	if possible jump to next butten, otherwise jump to next element
	//	CTRL-RIGHT: move selected value to the right
	//	INS:	insert value after selected value
	//	DEL:	remove selected value
	//
	// listfunc: array function(); should return array of possible values
	// func: array function(array);
	// rbfunc: function(x, y, valuenr, hoverx, hovery, hoverw, hoverh);

	this.AddKeyWords = function (x, y, w, skip, font, listfunc, func, rbfunc) {this.list.push(new FormKeyWords(x, y, w, skip, font, listfunc, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddKeyWordsH(x, y, w, skip, font[, listfunc, func, rbfunc])
	//--------------------
	// Draws single line showing a list of key words which can be set or not set.
	//--------------------
	// Using keyboard:
	//	ENTER: add or remove selected value
	//	END: jump to last value
	//	HOME: jump to first value
	//	LEFT: if possible select previous value, otherwise jump to previous element
	//	SHIFT-TAB, UP: jump to previous element
	//	RIGHT: if possible select next value, otherwise jump to next element
	//	TAB, DOWN: jump to next element
	//
	// listfunc: array function(); should return array of possible values
	// func: array function(array);
	// rbfunc: function(x, y, valuenr, hoverx, hovery, hoverw, hoverh);

	this.AddKeyWordsH = function (x, y, w, skip, font, listfunc, func, rbfunc) {this.list.push(new FormKeyWordsH(x, y, w, skip, font, listfunc, func, rbfunc))}

	//--------------------
	// function FormsAndButtons.AddKeyWordsHS(x, y, w, skip, font[, listfunc, func, rbfunc])
	//--------------------
	// Draws single line showing a list of key words of which only one can be set.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, UP: jump to previous element
	//	TAB, DOWN: jump to next element
	//	ENTER: add or remove selected value
	//	END: jump to last value
	//	HOME: jump to first value
	//	LEFT: if possible select previous value, otherwise jump to previous element
	//	RIGHT: if possible select next value, otherwise jump to next element
	//
	// listfunc: array function(); should return array of possible values
	// func:	string function(string);
	// rbfunc: funcion(x, y, valuenr, hoverx, hovery, hoverw, hoverh);

	this.AddKeyWordsHS = function (x, y, w, skip, font, listfunc, func, rbfunc) {this.list.push(new FormKeyWordsHS(x, y, w, skip, font, listfunc, func, rbfunc))}

	//--------------------
	// function FormsAndButtos.AddRating(x, y, skip, font, max[, func, rbfunc])
	//--------------------
	// Draws line of stars to edit rating.
	//--------------------
	// Using keyboard:
	//	SHIFT-TAB, UP: jump to previous element
	//	TAB, DOWN: jump to next element
	//	ENTER: set selected rating or if this rating is set, set rating to 0
	//	END: jump to maximum
	//	HOME: jump to 1
	//	LEFT: if possible select previous rating, otherwise jump to previous element
	//	RIGHT: if possible select next rating, otherwise jump to next element
	//
	// max: maximum rating
	// func: int function(int);
	// rbfunc: funcion(x, y, valuenr, hoverx, hovery, hoverw, hoverh);

	this.AddRating = function (x, y, skip, font, max, func, rbfunc) {this.list.push(new FormRating(x, y, skip, font, max, func, rbfunc))}

	//--------------------
	// function FormDrawString(x, y, font, colour, str)
	//--------------------
	// Draws string.

	this.AddString = function (x, y, font, colour, str) {this.list.push(new FormString(x, y, font, colour, str))}

	//--------------------
	// These functions work similarly to the ones above
	//--------------------

	this.PopUpAddIButton = function (x, y, w, h, skip, imgsrc, func, rbfunc) {this.popuplist.push(new FormIButton("px + " + x, "py + " + y, w, h, skip, imgsrc, func, rbfunc))}
	this.PopUpAddTButton = function (x, y, w, h, skip, text, font, func, rbfunc) {this.popuplist.push(new FormTButton("px + " + x, "py + " + y, w, h, skip, text, font, func, rbfunc))}
	this.PopUpAddTButtonAlt = function (x, y, w, h, skip, text, font, func, rbfunc) {this.popuplist.push(new FormTButtonAlt("px + " + x, "py + " + y, w, h, skip, text, font, func, rbfunc))}
	this.PopUpAddTextBox = function (x, y, w, h, skip, font, func, rbfunc) {this.popuplist.push(new FormTextBox("px + " + x, "py + " + y, w, h, skip, font, func, rbfunc))}
	this.PopUpAddText = function (x, y, w, h, skip, font, func, rbfunc) {this.popuplist.push(new FormText("px + " + x, "py + " + y, w, h, skip, font, func, rbfunc))}
	this.PopUpAddTextInput = function (x, y, w, skip, font, func, rbfunc) {this.popuplist.push(new FormTextInput("px + " + x, "py + " + y, w, skip, font, func, rbfunc))}
	this.PopUpAddTextLine = function (x, y, w, skip, font, func, rbfunc) {this.popuplist.push(new FormTextLine("px + " + x, "py + " + y, w, skip, font, func, rbfunc))}
	this.PopUpAddMVTextInput = function (x, y, w, skip, font, func, rbfunc) {this.popuplist.push(new FormMVTextInput("px + " + x, "py + " + y, w, skip, font, func, rbfunc))}
	this.PopUpAddMVSTextInput = function (x, y, w, skip, font, func, rbfunc) {this.popuplist.push(new FormMVSTextInput("px + " + x, "py + " + y, w, skip, font, func, rbfunc))}
	this.PopUpAddMVTextLine = function (x, y, w, skip, font, func, rbfunc) {this.popuplist.push(new FormMVTextLine("px + " + x, "py + " + y, w, skip, font, func, rbfunc))}
	this.PopUpAddKeyWords = function (x, y, w, skip, font, listfunc, func, rbfunc) {this.popuplist.push(new FormKeyWords("px + " + x, "py + " + y, w, skip, font, listfunc, func, rbfunc))}
	this.PopUpAddKeyWordsH = function (x, y, w, skip, font, listfunc, func, rbfunc) {this.popuplist.push(new FormKeyWordsH("px + " + x, "py + " + y, w, skip, font, listfunc, func, rbfunc))}
	this.PopUpAddKeyWordsHS = function (x, y, w, skip, font, listfunc, func, rbfunc) {this.popuplist.push(new FormKeyWordsHS("px + " + x, "py + " + y, w, skip, font, listfunc, func, rbfunc))}
	this.PopUpAddRating = function (x, y, skip, font, max, func, rbfunc) {this.popuplist.push(new FormRating("px + " + x, "py + " + y, skip, font, max, func, rbfunc))}
	this.PopUpAddString = function (x, y, font, colour, str) {this.popuplist.push(new FormString("px + " + x, "py + " + y, font, colour, str))}

	//--------------------
	// function FormsAndButtons.Reload()	
	//--------------------
	// Makes all elements reload values

	this.Reload = function () {
		SimpleInput.reload = true;
		for (var i in this.list) this.list[i].reload = true;
		for (var i in this.popuplist) this.popuplist[i].reload = true;
		window.Repaint();
	}
	
	//--------------------
	// function FormsAndButtons.SetPopUp(x, y, w, h)
	//--------------------
	// Sets coordinates and size of pop-up
	//
	// if x == 0 pop-up is centred horizontally
	// if y == 0 pop-up is centred vertically
	// if w == 0 width is set to maximum
	// if h == 0 height is set to maximum

	this.SetPopUp = function (x, y, w, h) {
		if (arguments.length > 0) {
			this.popupx = x;
			this.popupy = y;
			this.popupw = w;
			this.popuph = h;
		}
		px = eval(this.popupx);
		py = eval(this.popupy);
		pw = eval(this.popupw);
		ph = eval(this.popuph);
		if (pw <= 0) pw = ww;
		if (ph <= 0) ph = wh;
		if (px < 0) px = (ww - pw) / 2
		if (py < 0) py = (ww - ph) / 2
	}

	//--------------------
	// function FormsAndButtons.OpenPopUp()
	//--------------------
	// Opens pop-up

	this.OpenPopUp = function () {
		this.popup = true;
		for (var i in this.list) {
			this.list[i].on_activate(false);
		}
		FormsAndButtons.ActivateNext();
		window.Repaint();
	}

	//--------------------
	// function FormsAndButtons.ClosePopUp()
	//--------------------
	// Closes pop-up

	this.ClosePopUp = function () {
		this.popup = false;
		for (var i in this.popuplist) {
			this.popuplist[i].on_activate(false);
		}
		FormsAndButtons.ActivateNext();
		window.Repaint();
	}

	//--------------------
	// function FormsAndButtons.EmptyPopUp()	
	//--------------------
	// Closes pop-up and disposes all its elements

	this.EmptyPopUp = function () {
		for (var i in this.popuplist) {
			this.popuplist[i].on_activate(false);
			this.popuplist[i].Dispose();
		}
		this.popuplist = [];
		window.Repaint();
	}

	//--------------------
	// function FormsAndButtons.ActivateNext()
	//--------------------
	// Activates next element with skip variable set to false

	this.ActivateNext = function () {
		var first = null;
		var active = null;
		var next = null;
		var objects = this.popup ? this.popuplist : this.list;
		for (var i in objects) {
			if (!objects[i].skip) {
				if (!first) first = i;
				if (active && !next) next = i;
				if (objects[i].active) active = i;
			}
		}
		if (!next) next = first;
		for (var i in objects) {
			objects[i].on_activate(false);
		}
		this.activeobj = next;
		if (next) objects[next].on_activate(true);
	}

	//--------------------
	// function FormsAndButtons.ActivatePrevious()
	//--------------------
	// Activates previous element with skip variable set to false

	this.ActivatePrevious = function () {
		var first = null;
		var active = null;
		var previous = null;
		var tmp = null;
		var objects = this.popup ? this.popuplist : this.list;
		for (var i in objects) {
			if (!objects[i].skip) {
				if (!first) first = i;
				if (objects[i].active) {
					active = i;
					if (tmp) previous = tmp;
				}
				tmp = i;
			}
		}
		if (!previous) previous = tmp;
		for (var i in objects) {
			objects[i].on_activate(false);
		}
		this.activeobj = previous;
		if (previous) objects[previous].on_activate(true);
	}

	//--------------------
	// function FormsAndButtons.Dispose()
	//--------------------
	// Closes and disposes all elements
	
	this.Dispose = function () {
		SimpleInput.Close();
		for (var i in this.list) {
			this.list[i].on_activate(false);
			this.list[i].Dispose();
		}
		for (var i in this.popuplist) {
			this.popuplist[i].on_activate(false);
			this.popuplist[i].Dispose();
		}
		this.list = [];
		this.popuplist = [];
		window.Repaint();
	}

	this.list = [];
	this.popuplist = [];
	this.activeobj = null;
	this.popup = false;
	this.popupx = 0;
	this.popupy = 0;
	this.popuph = 0;
	this.popupw = 0;
	this.resize = false;
	this.cvisible = true;
	this.crpx = 0;
	this.crpy = 0;
	this.crpw = 0;
	this.crph = 0;
	this.CopyBuffer = "";
}

SimpleInput = new FormSimpleInput;

FormSimpleInput.prototype = new FormProto;
FormSimpleInput.prototype.constructor = FormSimpleInput;

function FormSimpleInput() {

	//--------------------
	// function SimpleInput.Open(str, x, y, w, font[, func, rbfunc])
	//--------------------
	// Opens an text input line, indepentend from the FormsAndButtons structure
	//	
	// Using keyboard :
	//	BKSP, END, HOME, LEFT, RIGHT, INS, DEL, CTRL-A, CTRL-C, CTRL-V, CTRL-X
	//		as expected in a text editor (hold SHIFT to select)
	//	ENTER:	exit edit mode and send value if changed
	//	ESC: exit edit mode and send value if changed
	//	CTRL-Z: undo change
	//
	// func: string function(string);
	// rbfunc: function(x, y, hoverx, hovery, hoverw, hoverh);

	this.Open = function (str, x, y, w, font, func, rbfunc) {
		this.active = true;
		for (var i in intobjects) {
			intobjects[i].on_activate(false);
		}
		for (var i in popupobjects) {
			popupobjects[i].on_activate(false);
		}
		this.Text = str;
		this.SText = str;
		this.rx = x;
		this.x = Math.max(0, eval(x));
		this.ry = y;
		this.y = Math.max(0, eval(y));
		this.rw = w;
		this.w = Math.max(0, eval(w));
		this.Font = font;
		this.func = func;
		this.rbfunc = rbfunc;
		var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
		this.THeight = arr[3] - arr[1];
		this.CPos = 0;
		this.select = false;
		this.Offset = 0;
		this.CalcText();
		window.Repaint();
	}

	//--------------------
	// function SimpleInput.Close()
	//--------------------
	// Closes the text input line
	
	this.Close = function () {
		if (this.active) {
			this.active = false;
			if (this.Text != this.SText) if (this.func) {
				this.func(this.Text);
				FormsAndButtons.Reload();
			}
		}
	}
	
	this.rx = 0;
	this.x = 0;
	this.ry = 0;
	this.y = 0;
	this.rw = 0;
	this.w = 0;
	this.Text = "";
	this.SText = "";
	this.Font = null;
	this.func = null;
	this.rbfunc = null;
	this.Offset = 0;
	this.TLength = null;
	this.TWidth = null;
	this.THeight = null;
	this.CPos = 0;
	this.SPos = 0;
	this.Cx = 0;
	this.THeight = 0;
	this.insert = false;
	this.select = false;

	this.resize = function () {
		this.x = Math.max(0, eval(this.rx));
		this.y = Math.max(0, eval(this.ry));
		this.w = Math.max(0, eval(this.rw));
	}

	this.draw = function (gr) {
		if (this.active) {
			gr.FillSolidRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, editbackcolour);
			gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
			this.ForceVisible();
			if (this.select) this.DrawSelection(gr);
			gr.GdiDrawText(this.Text.substr(this.Offset), this.Font, edittextcolour, this.x, this.y, this.w, this.THeight, DT = DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT);
			this.DrawCursor(gr);
		}
	}

	this.on_mousemove = function (x, y) {
		if (this.active) {
			var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
			if (hov != this.hover) {
				if (hov) {
					if (this.w < this.TWidth) {
						tooltip.Text = this.Text;
						tooltip.Activate();
					}
				} else tooltip.Deactivate();
				window.SetCursor(hov ? IDC_IBEAM : IDC_ARROW);
				this.hover = hov;
			}
			if (this.lbtnd) {
				var tmp = this.CPos;
				this.GetCPos(x);
				if (this.Cpos != this.SPos) this.select = true;
				if (tmp != this.CPos) ommRepaint = true;
			}
		}
	}

	this.on_mouse_lbtn_down = function (x, y) {
		if (this.active) {
			this.select = false;
			this.GetCPos(x);
			this.SPos = this.CPos;
			omldRepaint = true;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		return true;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.active && this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x, this.y, this.w, this.THeight);
			FormsAndButtons.Reload();
			return true;
		}
	}

	this.on_key = function (vkey, mask) {
		if (this.active) {
			if (mask == KMask.none) {
				switch (vkey) {
				case VK_BACK:
					if (this.select) {
						this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText();
					} else {
						if (this.CPos > 0) {
							this.Text = this.Text.slice(0, this.CPos - 1) + this.Text.slice(this.CPos);
							this.CPos--;
							this.CalcText();
						}
					}
					break;
				case VK_RETURN:
					this.Close();
					break;
				case VK_ESCAPE:
					this.Close();
					break;
				case VK_END:
					this.CPos = this.TLength;
					this.select = false;
					break;
				case VK_HOME:
					this.CPos = 0;
					this.Offset = 0;
					this.select = false;
					break;
				case VK_LEFT:
					if (this.select) {
						this.CPos = this.SelBegin;
						this.select = false;
					} else if (this.CPos > 0) this.CPos--;
					break;
				case VK_RIGHT:
					if (this.select) {
						this.CPos = this.SelEnd;
						this.select = false;
					} else if (this.CPos < this.TLength) this.CPos++;
					break;
				case VK_INSERT:
					this.insert = !this.insert;
					break;
				case VK_DELETE:
					if (this.select) {
						this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText();
					} else {
						this.Text = this.Text.slice(0, this.CPos) + this.Text.slice(this.CPos + 1);
						this.CalcText();
					}
					break;
				}
				okdRepaint = true;
			}
			if (mask == KMask.shift) {
				switch (vkey) {
				case VK_END:
					if (!this.select) {
						this.SPos = this.CPos;
						this.select = true;
					}
					this.CPos = this.TLength;
					break;
				case VK_HOME:
					if (!this.select) {
						this.SPos = this.CPos;
						this.select = true;
					}
					this.CPos = 0;
					this.Offset = 0;
					break;
				case VK_LEFT:
					if (!this.select) {
						this.SPos = this.CPos;
						this.select = true;
					}
					if (this.CPos > 0) this.CPos--;
					break;
				case VK_RIGHT:
					if (!this.select) {
						this.SPos = this.CPos;
						this.select = true;
					}
					if (this.CPos < this.TLength) this.CPos++;
					break;
				}
				okdRepaint = true;
			}
		}
	}

	this.on_char = function (code, mask) {
		if (this.active) {
			if (code == 1 && mask == KMask.ctrl) {
				// ctrl a
				this.SPos = 0;
				this.CPos = this.Text.length;
				this.select = true;
				ocRepaint = true;
			}
			if (code == 3 && this.select && mask == KMask.ctrl) {
				// ctrl c
				FormsAndButtons.CopyBuffer = this.Text.substr(this.SelBegin, this.SelEnd - this.SelBegin);
			}
			if (code == 22 && mask == KMask.ctrl) {
				// ctrl v
				if (FormsAndButtons.CopyBuffer.length > 0) {
					if (this.select) {
						var p1 = this.SelBegin;
						var p2 = this.SelEnd;
						this.CPos = this.SelBegin;
						this.select = false;
					} else {
						var p1 = this.CPos;
						var p2 = this.CPos;
						if (this.insert) p2 += FormsAndButtons.CopyBuffer.length;
						if (p2 > this.Text.length - 1) p2 = this.Text.length - 1;
					}
					this.Text = this.Text.slice(0, p1) + FormsAndButtons.CopyBuffer + this.Text.slice(p2);
					this.CPos += FormsAndButtons.CopyBuffer.length;
					this.select = false;
					this.CalcText();
					this.CalcOText();
					ocRepaint = true;
				}
			}
			if (code == 24 && this.select && mask == KMask.ctrl) {
				// ctrl x
				FormsAndButtons.CopyBuffer = this.Text.substr(this.SelBegin, this.SelEnd - this.SelBegin);
				this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
				this.CPos = this.SelBegin;
				this.select = false;
				this.CalcText();
				this.CalcOText();
				ocRepaint = true;
			}
			if (code == 9) {
				// tab
				this.Close();
			}
			if (code == 26) {
				// ctrl z
				this.Text = this.SText;
				this.CPos = this.TLength;
				this.select = false;
				this.CalcText();
				ocRepaint = true;
			}
			if (code > 31) {
				if (this.select) {
					var p1 = this.SelBegin;
					var p2 = this.SelEnd;
					this.CPos = this.SelBegin;
					this.select = false;
				} else {
					var p1 = this.CPos;
					var p2 = this.CPos;
					if (this.insert) p2++;
				}
				this.Text = this.Text.slice(0, p1) + String.fromCharCode(code) + this.Text.slice(p2);
				this.CPos++;
				this.CalcText();
				ocRepaint = true;
			}
		}
	}

	this.CalcText = function () {
		var arr = temp_gr.GdiDrawText(this.Text, this.Font, 0, 0, 0, 9999999, 9999999, DT_NOPREFIX | DT_CALCRECT).toArray();
		this.TLength = arr[4];
		this.TWidth = temp_gr.CalcTextWidth(this.Text, this.Font);
	}

	this.DrawSelection = function (gr) {
		var tmp = this.CPos > this.SPos;
		this.SelBegin = tmp ? this.SPos : this.CPos;
		this.SelEnd = tmp ? this.CPos : this.SPos;
		if (this.SelBegin < this.Offset) this.SelBegin = this.Offset;
		if (this.SelBegin != this.SelEnd) {
			var Sx1 = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.SelBegin - this.Offset), this.Font);
			var Sx2 = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.SelEnd - this.Offset), this.Font);
			gr.FillSolidRect(this.x + Sx1, this.y, Sx2 - Sx1, this.THeight, selectcolour);
		}
	}

	this.DrawCursor = function (gr) {
		if (this.insert) {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx + Math.round(0.3 * this.THeight);
			var y1 = this.y + this.THeight;
			var y2 = this.y + this.THeight;
			var lt = 2;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1 - 1;
			FormsAndButtons.crpw = x2 - x1 + 1;
			FormsAndButtons.crph = lt + 1;
		} else {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx;
			var y1 = this.y;
			var y2 = this.y + this.THeight;
			var lt = 1;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1;
			FormsAndButtons.crpw = lt;
			FormsAndButtons.crph = y2 - y1 + 1;
		}
		if (FormsAndButtons.cvisible) gr.DrawLine(x1, y1, x2, y2, lt, edittextcolour);
	}

	this.ForceVisible = function () {
		while (this.CPos < this.Offset) this.Offset--;
		this.Cx = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.CPos - this.Offset), this.Font);
		while (this.Cx > this.w) {
			this.Offset++;
			this.Cx = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.CPos - this.Offset), this.Font);
		}
	}

	this.GetCPos = function (x) {
		x -= this.x
		this.CPos = this.Offset;
		var trlen = temp_gr.CalcTextWidth(this.Text.substr(this.Offset), this.Font);
		if (x > trlen) this.CPos = this.TLength;
		else {
			var arr = temp_gr.GdiDrawText(this.Text.substr(this.Offset), this.Font, 0, 0, 0, this.w, this.THeight, DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			var tmp = x / trlen * arr[4];
			var td1 = x - temp_gr.CalcTextWidth(this.Text.substr(this.Offset, tmp), this.Font);
			var sign = td1 == 0 ? 1 : Math.abs(td1) / td1;
			var e = 0;
			do {
				tmp += sign;
				td2 = x - temp_gr.CalcTextWidth(this.Text.substr(this.Offset, tmp), this.Font);
				t = td1;
				td1 = td2;
				e++;
				if (e > 100) {
					fb.trace("ERROR SimpleInput.GetCPos");
					break
				}
			} while (Math.abs(td2) < Math.abs(t));
			this.CPos += Math.floor(tmp - sign);
		}
	}
}

//--------------------
// PROTOTYPE
//--------------------

function FormProto() {
	this.skip = true; // set to false to prevent activating
	this.hover = false;
	this.lbtnd = false; // is true when left mouse button is down
	this.rbtnd = false; // is true when right mouse button is down
	this.active = false;
	this.reload = false; // is true when asked for reload

	this.resize = function () {
		// called on resize
	}

	this.draw = function (gr) {
		// called on paint
	}

	this.on_mousemove = function (x, y) {
		// called on mouse move
		// should set this.hover
		// for repaint: set ommRepaint = true
		//	(do not set ommRepaint = false!!!)
	}

	this.on_mouse_lbtn_down = function (x, y) {
		// called on left mouse button down
		// for repaint: omldRepaint = true
		//	(do not set omldRepaint = false!!!)
	}

	this.on_mouse_lbtn_up = function (x, y) {
		// called on left mouse button up
		// for repaint: omluRepaint = true
		//	(do not set omluRepaint = false!!!)
		// return true to set as active
	}

	this.on_mouse_rbtn_down = function (x, y) {
		// called on right mouse button down
		// for repaint: omrdRepaint = true
		//	(do not set omrdRepaint = false!!!)
	}

	this.on_mouse_rbtn_up = function (x, y) {
		// called on right mouse button up
		// for repaint: omruRepaint = true
		//	(do not set omruRepaint = false!!!)
		// return true to set as active and to supress system context menu
	}

	this.on_mouse_wheel = function (step) {
		// called on mouse wheel
		// for repaint: omwRepaint = true
		//	(do not set omwRepaint = false!!!)
	}

	this.on_activate = function (act) {
		// called to activate (act == true) or deactivate (act == false)
		this.active = this.skip ? false : act;
	}

	this.on_key = function (vkey, mask) {
		// called on key down
		// for repaint: okdRepaint = true
		//	(do not set okdRepaint = false!!!)
		if (mask == KMask.none) {
			if (vkey == VK_RIGHT || vkey == VK_DOWN) {
				FormsAndButtons.ActivateNext();
				okdRepaint = true;
			}
			if (vkey == VK_LEFT || vkey == VK_UP) {
				FormsAndButtons.ActivatePrevious();
				okdRepaint = true;
			}
		}
	}

	this.on_char = function (code, mask) {
		// called on char
		// for repaint: ocRepaint = true
		//	(do not set ocRepaint = false!!!)
		if (code == 9) {
			// tab
			if (mask == KMask.none) FormsAndButtons.ActivateNext();
			if (mask == KMask.shift) FormsAndButtons.ActivatePrevious();
			ocRepaint = true;
		}
	}
	
	this.Dispose = function () {
		// called by FormsAndButtons.Dispose();
	}
}

//--------------------

FormIButton.prototype = new FormProto;
FormIButton.prototype.constructor = FormIButton;

function FormIButton(x, y, w, h, skip, imgsrc, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.h = Math.max(0, eval(h));
	this.skip = skip;
	this.imgsrc = imgsrc;
	this.img = null;
	this.check = null;
	this.tcheck = null;
	this.tactive = null;
	this.func = func;
	this.rbfunc = rbfunc;

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
		this.h = Math.max(0, eval(h));
	}

	this.draw = function (gr) {
		this.check = this.func ? this.func() : false;
		if (this.check != this.tcheck || this.active != this.tactive || !this.img) {
			this.tcheck = this.check;
			this.tactive = this.active;
			if (this.img) 
				this.img.Dispose();
			if (this.active) {
				this.img = this.check ? gdi.Image(imgsrc.activeon) : gdi.Image(imgsrc.active);
			}
			else {
				this.img = this.check ? gdi.Image(imgsrc.inactiveon) : gdi.Image(imgsrc.inactive);
			}
		}
		gr.DrawImage(this.img, this.x, this.y, this.w, this.h, 0, 0, this.img.Width, this.img.Height);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.h)) ? true : false;
		if (hov != this.hover) {
			window.SetCursor(hov ? IDC_HAND : IDC_ARROW);
			this.hover = hov;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			this.tactive = this.active;
			if (!this.skip) this.active = true;
			this.tcheck = this.check;
			this.check = this.check ? false : true;
			if (this.func) this.func(this.check);
			FormsAndButtons.Reload();
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x, this.y, this.w, this.h);
			FormsAndButtons.Reload();
			return true;
		}
	}

	this.on_activate = function (act) {
		this.tactive = this.active;
		this.active = this.skip ? false : act;
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			if (vkey == VK_RIGHT || vkey == VK_DOWN) {
				FormsAndButtons.ActivateNext();
				okdRepaint = true;
			}
			if (vkey == VK_LEFT || vkey == VK_UP) {
				FormsAndButtons.ActivatePrevious();
				okdRepaint = true;
			}
			if (vkey == VK_RETURN) {
				this.check = this.check ? false : true;
				if (this.func) this.func(this.check);
				okdRepaint = true;
			}
		}
	}

	this.Dispose = function () {
		if (this.img) this.img.Dispose();
	}
}

FormTButton.prototype = new FormProto;
FormTButton.prototype.constructor = FormTButton;

function FormTButton(x, y, w, h, skip, text, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.h = Math.max(0, eval(h));
	this.skip = skip;
	this.text = text;
	this.font = font;
	this.check = null;
	this.func = func;
	this.rbfunc = rbfunc;
	var arr = temp_gr.GdiDrawText(".", this.font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];
	this.TWidth = temp_gr.CalcTextWidth(this.text, this.font);

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
		this.h = Math.max(0, eval(h));
		if (this.w == 0) this.w = this.TWidth;
		else {
			this.x += marginx;
			this.w -= 2 * marginx;
		}
		if (this.h == 0) this.h = this.THeight;
		else {
			this.y += marginy;
			this.h -= 2 * marginy;
		}
	}

	this.draw = function (gr) {
		this.check = this.func ? this.func() : false;
		var ah = this.h > 0 ? marginy : this.h / 2 + marginy;
		var aw = this.w > 0 ? marginx : this.w / 2 + marginx;
		var arc = aw < ah ? aw : ah;
		var colour = this.check ? checkedcolour : backcolour;
		gr.FillRoundRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.h + 2 * marginy, arc, arc, colour);
		var colour = this.active ? activecolour : textcolour;
		gr.DrawRoundRect(this.x - marginx + 1, this.y - marginy + 1, this.w + 2 * marginx - 2, this.h + 2 * marginy - 2, arc, arc - 1, 2, colour);
		gr.GdiDrawText(this.text, this.font, textcolour, this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.h + 2 * marginy, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT | DT_NOPREFIX);
	}

	this.on_mousemove = function (x, y) {
		var hov = ((this.x - marginx) < x && x < (this.x + this.w + 2 * marginx) && (this.y - marginy) < y && y < (this.y + this.h + 2 * marginy)) ? true : false;
		if (hov != this.hover) {
			window.SetCursor(hov ? IDC_HAND : IDC_ARROW);
			this.hover = hov;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			this.check = this.check ? false : true;
			if (this.func) this.func(this.check);
			FormsAndButtons.Reload();
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.h + 2 * marginy);
			FormsAndButtons.Reload();
			return true;
		}
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			if (vkey == VK_RIGHT || vkey == VK_DOWN) {
				FormsAndButtons.ActivateNext();
				okdRepaint = true;
			}
			if (vkey == VK_LEFT || vkey == VK_UP) {
				FormsAndButtons.ActivatePrevious();
				okdRepaint = true;
			}
			if (vkey == VK_RETURN) {
				this.check = this.check ? false : true;
				if (this.func) this.func(this.check);
				okdRepaint = true;
			}
		}
	}
}

FormTButtonAlt.prototype = new FormProto;
FormTButtonAlt.prototype.constructor = FormTButtonAlt;

function FormTButtonAlt(x, y, skip, text, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.skip = skip;
	this.text = text;
	this.font = font;
	this.check = null;
	this.func = func;
	this.rbfunc = rbfunc;
	var arr = temp_gr.GdiDrawText(".", this.font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];
	this.TWidth = temp_gr.CalcTextWidth(this.text, this.font);

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
	}

	this.draw = function (gr) {
		this.check = this.func ? this.func() : false;
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.TWidth + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
		var colour = this.check ? textcolour : uncheckedcolour;
		gr.GdiDrawText(this.text, this.font, colour, this.x, this.y, this.TWidth, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < this.x + this.TWidth && this.y < y && y < this.y + this.THeight) ? true : false;
		if (hov != this.hover) {
			window.SetCursor(hov ? IDC_HAND : IDC_ARROW);
			this.hover = hov;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			this.check = this.check ? false : true;
			if (this.func) this.func(this.check);
			FormsAndButtons.Reload();
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x, this.y, this.TWidth, this.THeight);
			FormsAndButtons.Reload();
			return true;
		}
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			if (vkey == VK_RIGHT || vkey == VK_DOWN) {
				FormsAndButtons.ActivateNext();
				okdRepaint = true;
			}
			if (vkey == VK_LEFT || vkey == VK_UP) {
				FormsAndButtons.ActivatePrevious();
				okdRepaint = true;
			}
			if (vkey == VK_RETURN) {
				this.check = this.check ? false : true;
				if (this.func) this.func(this.check);
				okdRepaint = true;
			}
		}
	}
}

FormTextBox.prototype = new FormProto;
FormTextBox.prototype.constructor = FormTextBox;

function FormTextBox(x, y, w, h, skip, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.h = Math.max(0, eval(h));
	this.skip = skip;
	this.Font = font;
	this.func = func;
	this.rbfunc = rbfunc;
	this.Text = "";
	this.SText = "";
	this.Offset = 0;
	this.TRows = null;
	this.Rows = null;
	this.RLength = [];
	this.RBegin = [];
	this.THeight = null;
	this.OBegin = 0;
	this.OLength = 0;
	this.CPos = 0;
	this.SPos = 0;
	this.SelBegin = 0;
	this.SelEnd = 0;
	this.CRow = 0;
	this.CRowT = 0;
	this.Cx = 0;
	this.Cy = 0;
	this.ScrY = 0;
	this.SOffset = 0;
	this.hov = false;
	this.HovU = false;
	this.HovD = false;
	this.edit = false;
	this.insert = false;
	this.select = false;
	this.noenter = false;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
		this.h = Math.max(0, eval(h));
		this.CalcText();
		this.CalcOText();
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var str = this.func();
				if (str != this.SText) {
					this.Text = str;
					this.SText = str;
					this.Offset = 0;
					this.edit = false;
					this.select = false;
					this.CalcText();
					this.CalcOText();
				}
			}
			this.reload = false;
		}
		if (this.edit) {
			gr.FillSolidRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.h + 2 * marginy, editbackcolour);
			this.ForceVisible();
			if (this.select) this.DrawSelection(gr);
			var colour = edittextcolour;
		} else var colour = textcolour;
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.h + 2 * marginy, 1, activecolour);
		var arr = gr.GdiDrawText(this.Text.substr(this.OBegin), this.Font, colour, this.x, this.y, this.w, this.h, DT_LEFT | DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
		if (this.edit) {
			this.DrawCursor(gr);
		} else if (this.hover) {
			var tfont = gdi.Font("Tahoma", 8);
			if (this.Offset > 0) {
				gr.FillSolidRect(this.x + this.w + marginx - 4, this.y, 7, 7, backcolour);
				gr.DrawRect(this.x + this.w + marginx - 4, this.y, 7, 7, 1, textcolour);
				gr.GdiDrawText("+", tfont, textcolour, this.x + this.w + marginx - 4, this.y - 1, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
			}
			if (this.Offset < this.TRows - this.Rows) {
				gr.FillSolidRect(this.x + this.w + marginx - 4, this.y + this.h - 8, 7, 7, backcolour);
				gr.DrawRect(this.x + this.w + marginx - 4, this.y + this.h - 8, 7, 7, 1, textcolour);
				gr.GdiDrawText("-", tfont, textcolour, this.x + this.w + marginx - 4, this.y + this.h - 9, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
			}
		}
	}

	this.on_mousemove = function (x, y) {
		var T = (this.x < x && x < this.x + this.w + marginx + 3 && this.y < y && y < this.y + this.h) ? true : false;
		var hov = (this.x < x && x < this.x + this.w && this.y < y && y < this.y + this.h) ? true : false;
		var U = (this.x + this.w < x && x < this.x + this.w + marginx + 3 && this.y < y && y < this.y + 8 && this.Offset > 0) ? true : false;
		var D = (this.x + this.w < x && x < this.x + this.w + marginx + 3 && this.y + this.h - 8 < y && y < this.y + this.h && this.Offset < this.TRows - this.Rows) ? true : false;
		if (T != this.hover || hov != this.hov || U != this.HovU || D != this.HovD) {
			var tcur = IDC_ARROW;
			if (hov) tcur = IDC_IBEAM;
			if (U) tcur = IDC_HAND;
			if (D) tcur = IDC_HAND;
			window.SetCursor(tcur);
			this.hover = T;
			this.hov = hov;
			this.HovU = U;
			this.HovD = D;
			ommRepaint = true;
		}
		if (this.lbtnd) {
			var tmp = this.CPos;
			this.GetCPos(x, y);
			if (this.Cpos != this.SPos) this.select = true;
			if (tmp != this.CPos) ommRepaint = true;
		}
		if (this.rbtnd) {
			var TOffset = this.SOffset + Math.ceil((this.ScrY - y) / this.THeight);
			TOffset = Math.min(TOffset, this.TRows - this.Rows);
			TOffset = Math.max(TOffset, 0);
			if (TOffset != this.Offset) {
				this.Offset = TOffset;
				ommRepaint = true;
				this.CalcOText();
			}
		}
	}

	this.on_mouse_lbtn_down = function (x, y) {
		if (this.hov) {
			if (!this.skip) this.active = true;
			this.edit = true;
			this.select = false;
			this.GetCPos(x, y);
			this.SPos = this.CPos;
			omldRepaint = true;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.HovU) {
			if (this.Scroll(1, false)) omluRepaint = true;
		}
		if (this.HovD) {
			if (this.Scroll(-1, false)) omluRepaint = true;
		}
		return true;
	}

	this.on_mouse_rbtn_down = function (x, y) {
		this.ScrY = y;
		this.SOffset = this.Offset;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x, this.y, this.w, this.h);
			FormsAndButtons.Reload();
		}
		return true;
	}

	this.on_mouse_wheel = function (step) {
		if (this.hover) {
			if (this.Scroll(step, false)) omwRepaint = true;
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		if (!act && this.edit) {
			this.edit = false;
			if (this.Text != this.SText) if (this.func) this.func(this.Text);
		}
		this.CPos = 0;
		this.CRow = 0;
		this.select = false;
		this.CalcText();
		this.CalcOText();
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_BACK:
				if (this.edit) {
					if (this.select) {
						this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText();
						this.CalcOText();
					} else {
						if (this.CPos > 0) {
							this.Text = this.Text.slice(0, this.CPos - 1) + this.Text.slice(this.CPos);
							this.CPos--;
							this.CalcText();
							this.CalcOText();
						}
					}
				}
				break;
			case VK_ESCAPE:
				if (this.edit) {
					this.edit = false;
					okdRepaint = true;
					if (this.Text != this.SText) if (this.func) this.func(this.Text);
				}
				break;
			case VK_PGUP:
				if (this.Scroll(1, true)) okdRepaint = true;
				this.select = false;
				break;
			case VK_PGDN:
				if (this.Scroll(-1, true)) okdRepaint = true;
				this.select = false;
				break;
			case VK_END:
				this.CPos = this.RBegin[this.CRow] + this.RLength[this.CRow];
				this.select = false;
				break;
			case VK_HOME:
				this.CPos = this.RBegin[this.CRow];
				this.select = false;
				break;
			case VK_LEFT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelBegin;
						this.select = false;
					} else if (this.CPos > 0) this.CPos--;
				} else {
					FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_UP:
				if (this.edit) {
					if (this.CRow == this.Offset) {
						if (this.Scroll(1, false)) {
							this.GetCPos(this.Cx + this.x, this.Cy + this.y);
						} else this.CPos = this.RBegin[this.CRow];
					} else this.GetCPos(this.Cx + this.x, this.Cy + this.y - this.THeight);
					this.select = false;
				} else {
					if (!this.Scroll(1, false)) FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_RIGHT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelEnd;
						this.select = false;
					} else if (this.CPos < this.Text.length) this.CPos++;
				} else {
				FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_DOWN:
				if (this.edit) {
					if (this.CRow == this.Offset + this.Rows - 1) {
						if (this.Scroll(-1, false)) {
							this.GetCPos(this.Cx + this.x, this.Cy + this.y);
						} else this.CPos = this.RBegin[this.CRow] + this.RLength[this.CRow];
					} else {
						this.GetCPos(this.Cx + this.x, this.Cy + this.y + this.THeight);
						this.CRowT++;
					}
					this.select = false;
				} else {
					if (this.Scroll(-1, false)) FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_INSERT:
				this.insert = !this.insert;
				break;
			case VK_DELETE:
				if (this.edit) {
					if (this.select) {
						this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText();
						this.CalcOText();
					} else {
						this.Text = this.Text.slice(0, this.CPos) + this.Text.slice(this.CPos + 1);
						this.CalcText();
						this.CalcOText();
					}
				}
				break;
			case VK_RETURN:
				if (!this.edit) {
					this.Offset = 0;
					this.CPos = 0;
					this.edit = true;
					this.select = false;
					this.noenter = true;
				} else {
					this.noenter = false;
				}
				break;
			}
			if (this.edit) okdRepaint = true;
		}
		if (mask == KMask.shift && this.edit) {
			switch (vkey) {
			case VK_PGUP:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.Scroll(1, true);
				break;
			case VK_PGDN:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.Scroll(-1, true);
				break;
			case VK_END:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = this.RBegin[this.CRow] + this.RLength[this.CRow];
				break;
			case VK_HOME:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = this.RBegin[this.CRow];
				break;
			case VK_LEFT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos > 0) this.CPos--;
				break;
			case VK_UP:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CRow == this.Offset) {
					if (this.Scroll(1, false)) {
						this.GetCPos(this.Cx + this.x, this.Cy + this.y);
					} else this.CPos = this.RBegin[this.CRow];
				} else this.GetCPos(this.Cx + this.x, this.Cy + this.y - this.THeight);
				break;
			case VK_RIGHT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos < this.Text.length) this.CPos++;
				break;
			case VK_DOWN:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CRow == this.Offset + this.Rows - 1) {
					if (this.Scroll(-1, false)) {
						this.GetCPos(this.Cx + this.x, this.Cy + this.y);
					} else this.CPos = this.RBegin[this.CRow] + this.RLength[this.CRow];
				} else {
					this.GetCPos(this.Cx + this.x, this.Cy + this.y + this.THeight);
					this.CRowT++;
				}
				break;
			}
			okdRepaint = true;
		}
	}

	this.on_char = function (code, mask) {
		if (code == 1 && this.edit && mask == KMask.ctrl) {
			// ctrl a
			if (this.edit) {
				this.Scroll(-1, true);
				this.SPos = 0;
				this.CPos = this.Text.length;
				this.CRow = this.TRows - 1;
				this.select = true;
				ocRepaint = true;
			}
		}
		if (code == 3 && this.edit && mask == KMask.ctrl) {
			// ctrl c
			if (this.edit && this.select) {
				FormsAndButtons.CopyBuffer = this.Text.substr(this.SelBegin, this.SelEnd - this.SelBegin);
			}
		}
		if (code == 22 && this.edit && mask == KMask.ctrl) {
			// ctrl v
			if (this.edit && FormsAndButtons.CopyBuffer.length > 0) {
				if (this.select) {
					var p1 = this.SelBegin;
					var p2 = this.SelEnd;
					this.CPos = this.SelBegin;
					this.select = false;
				} else {
					var p1 = this.CPos;
					var p2 = this.CPos;
					if (this.insert) p2 += FormsAndButtons.CopyBuffer.length;
					if (p2 > this.Text.length) p2 = this.Text.length;
				}
				this.Text = this.Text.slice(0, p1) + FormsAndButtons.CopyBuffer + this.Text.slice(p2);
				this.CPos += FormsAndButtons.CopyBuffer.length;
				this.select = false;
				this.CalcText();
				this.CalcOText();
				ocRepaint = true;
			}
		}
		if (code == 24 && this.edit && mask == KMask.ctrl) {
			// ctrl x
			if (this.edit && this.select) {
				FormsAndButtons.CopyBuffer = this.Text.substr(this.SelBegin, this.SelEnd - this.SelBegin);
				this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
				this.CPos = this.SelBegin;
				this.select = false;
				this.CalcText();
				this.CalcOText();
				ocRepaint = true;
			}
		}
		if (code == 9) {
			// tab
			if (this.edit) {
				this.edit = false;
				if (this.Text != this.SText) if (this.func) this.func(this.Text);
			}
			if (mask == KMask.none) FormsAndButtons.ActivateNext();
			if (mask == KMask.shift) FormsAndButtons.ActivatePrevious();
			ocRepaint = true;
		}
		if (code == 26 && this.edit && mask == KMask.ctrl) {
			// ctrl z
			this.Text = this.SText;
			this.CPos = 0;
			this.CRow = 0;
			this.Offset = 0;
			this.select = false;
			this.CalcText();
			this.CalcOText();
			this.Scroll(1, true);
			ocRepaint = true;
		}
		if (((code == 13 && !this.noenter) || code > 31) && this.edit) {
			if (this.select) {
				var p1 = this.SelBegin;
				var p2 = this.SelEnd;
				this.CPos = this.SelBegin;
				this.select = false;
			} else {
				var p1 = this.CPos;
				var p2 = this.CPos;
				if (this.insert) p2++;
			}
			this.Text = this.Text.slice(0, p1) + String.fromCharCode(code) + this.Text.slice(p2);
			this.CPos++;
			this.CalcText();
			this.CalcOText();
			ocRepaint = true;
		}
	}
	
	this.CalcText = function () {
		if (this.Text.length == 0) {
			this.TRows = 1;
		} else {
			var arr = temp_gr.GdiDrawText(this.Text, this.Font, 0, 0, 0, this.w, 9999999, DT_VCENTER | DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			this.TRows = (arr[3] - arr[1]) / this.THeight;
			if (this.Text.slice(-1) == "\r") this.TRows++;
		}
		this.Rows = Math.floor(this.h / this.THeight);
	}

	this.CalcOText = function () {
		this.Offset = Math.max(this.Offset, 0);
		if (this.Offset > 0) {
			var arr = temp_gr.GdiDrawText(this.Text, this.Font, 0, 0, 0, this.w, this.Offset * this.THeight, DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			this.OBegin = arr[4];
		} else this.OBegin = 0;
		var tmppos = 0;
		this.RLength = [];
		for (var i = 0; i < this.TRows; i++) {
			this.RBegin[i] = tmppos;
			var arr = temp_gr.GdiDrawText(this.Text.substr(tmppos), this.Font, 0, 0, 0, this.w, this.THeight, DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			tmppos += arr[4];
			this.RLength[i] = arr[4];
			if ((this.Text.substr(tmppos - 1, 1) == " " || this.Text.substr(tmppos - 1, 1) == "\r") && i < this.TRows - 1) this.RLength[i]--;
		}
		this.OLength = this.TRows > this.Rows + this.Offset ? this.RBegin[this.Offset + this.Rows - 1] + this.RLength[this.Offset + this.Rows - 1] - this.OBegin : this.Text.length - this.OBegin;
	}

	this.Scroll = function (n, end) {
		if (end) {
			if (n > 0) {
				this.CPos = 0;
				this.CRow = 0;
				var TOffset = 0;
			} else {
				this.CPos = this.Text.length;
				this.CRow = this.TRows - 0;
				var TOffset = this.TRows - this.Rows;
			}
		}
		else var TOffset = this.Offset - n;
		TOffset = Math.min(TOffset, this.TRows - this.Rows);
		TOffset = Math.max(TOffset, 0);
		if (TOffset != this.Offset) {
			this.Offset = TOffset;
			this.CalcOText();
			return true
		} else return false
	}

	this.DrawSelection = function (gr) {
		this.SelBegin = Math.min(this.CPos, this.SPos);
		this.SelEnd = Math.max(this.CPos, this.SPos);
		var TSelBegin = Math.max(this.SelBegin, this.OBegin);
		var TSelEnd = Math.min(this.SelEnd, this.OBegin + this.OLength);
		if (TSelBegin != TSelEnd) {
			for (var i = 0; i < this.TRows; i++) {
				if (TSelBegin < this.RBegin[i + 1]) break;
				if (TSelBegin == this.RBegin[i] + this.RLength[i]) break;
			}
			var SRow1 = i < this.TRows ? i : this.TRows - 1;
			var Sx1 = temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[SRow1], TSelBegin - this.RBegin[SRow1]), this.Font);
			var Sy1 = (SRow1 - this.Offset) * this.THeight;
			for (var i = 0; i < this.TRows; i++) {
				if (TSelEnd < this.RBegin[i + 1]) break;
				if (TSelEnd == this.RBegin[i] + this.RLength[i]) break;
			}
			var SRow2 = i < this.TRows ? i : this.TRows - 1;
			var Sx2 = temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[SRow2], TSelEnd - this.RBegin[SRow2]), this.Font);
			var Sy2 = (SRow2 - this.Offset) * this.THeight;
			if (SRow1 == SRow2) gr.FillSolidRect(this.x + Sx1, this.y + Sy1, Sx2 - Sx1, this.THeight, selectcolour);
			if (SRow1 < SRow2) {
				gr.FillSolidRect(this.x + Sx1, this.y + Sy1, temp_gr.CalcTextWidth(this.Text.substr(TSelBegin, this.RBegin[SRow1] + this.RLength[SRow1] - TSelBegin), this.Font), this.THeight, selectcolour);
				for (var i = SRow1 + 1; i < SRow2; i++) {
					gr.FillSolidRect(this.x, this.y + (i - this.Offset) * this.THeight, temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[i], this.RLength[i]), this.Font), this.THeight, selectcolour);
				}
				gr.FillSolidRect(this.x, this.y + Sy2, temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[SRow2], TSelEnd - this.RBegin[SRow2]), this.Font), this.THeight, selectcolour);
			}
		}
	}

	this.DrawCursor = function (gr) {
		for (var i = 0; i < this.TRows; i++) {
			if (this.CPos < this.RBegin[i + 1]) break;
			if (this.CPos == this.RBegin[i] + this.RLength[i] && ((this.CRowT == i && this.RLength[i] > 0) || this.RLength[i] == 0)) break;
		}
		this.CRow = i < this.TRows ? i : this.TRows - 1;
		this.CRowT = this.CRow;
		if (this.CPos >= this.OBegin && this.CPos <= this.OBegin + this.OLength) {
			this.Cx = temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[this.CRow], this.CPos - this.RBegin[this.CRow]), this.Font);
			this.Cy = (this.CRow - this.Offset) * this.THeight;
			if (this.insert) {
				var x1 = this.x + this.Cx;
				var x2 = this.x + this.Cx + Math.round(0.3 * this.THeight);
				var y1 = this.y + this.Cy + this.THeight;
				var y2 = this.y + this.Cy + this.THeight;
				var lt = 2;
				FormsAndButtons.crpx = x1;
				FormsAndButtons.crpy = y1 - 1;
				FormsAndButtons.crpw = x2 - x1 + 1;
				FormsAndButtons.crph = lt + 1;
			} else {
				var x1 = this.x + this.Cx;
				var x2 = this.x + this.Cx;
				var y1 = this.y + this.Cy;
				var y2 = this.y + this.Cy + this.THeight;
				var lt = 1;
				FormsAndButtons.crpx = x1;
				FormsAndButtons.crpy = y1;
				FormsAndButtons.crpw = lt;
				FormsAndButtons.crph = y2 - y1 + 1;
			}
			if (FormsAndButtons.cvisible) gr.DrawLine(x1, y1, x2, y2, lt, edittextcolour);
		}
	}

	this.ForceVisible = function () {
		var e = 0;
		while (this.CPos < this.OBegin || (this.TRows <= this.Rows && this.Offset > 0)) {
			this.Scroll(1, false);
			e++;
			if (e > 100) {
				fb.trace("ERROR FormTextBox.ForceVisible 1");
				break
			}
		}
		var e = 0;
		while (this.CPos > this.OBegin + this.OLength) {
			this.Scroll(-1,false)
			e++;
			if (e > 100) {
				fb.trace("ERROR FormTextBox.ForceVisible 2");
				break
			}
		}
	}

	this.GetCPos = function (x, y) {
		x -= this.x;
		y -= this.y;
		this.CRow = Math.floor(y / this.THeight) + this.Offset;
		if (this.CRow > this.TRows - 1) {
			this.CRow = this.TRows - 1;
			this.CPos = this.Text.length;
		} else {
			var trlen = temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[this.CRow], this.RLength[this.CRow]), this.Font);
			if (x > trlen || trlen == 0) {
				this.CPos = this.RBegin[this.CRow] + this.RLength[this.CRow];
			}
			else {
				var tmp = x / trlen * this.RLength[this.CRow];
				var td1 = x - temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[this.CRow], tmp), this.Font);
				var sign = td1 == 0 ? 1 : Math.abs(td1) / td1;
				var e = 0;
				do {
					tmp += sign;
					td2 = x - temp_gr.CalcTextWidth(this.Text.substr(this.RBegin[this.CRow], tmp), this.Font);
					t = td1;
					td1 = td2;
					e++;
					if (e > 100) {
						fb.trace("ERROR FormTextBox.GetCPos");
						break
				}
				} while (Math.abs(td2) < Math.abs(t));
				this.CPos = Math.floor(this.RBegin[this.CRow] + tmp - sign);
			}
		}
	this.CRowT = this.CRow;
	}
}

FormText.prototype = new FormProto;
FormText.prototype.constructor = FormText;

function FormText(x, y, w, h, skip, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.h = Math.max(0, eval(h));
	this.skip = skip;
	this.Font = font;
	this.func = func;
	this.rbfunc = rbfunc;
	this.Text = "";
	this.SText = "";
	this.Offset = 0;
	this.TRows = null;
	this.Rows = null;
	this.THeight = null;
	this.OBegin = 0;
	this.ScrY = 0;
	this.SOffset = 0;
	this.hov = false;
	this.HovU = false;
	this.HovD = false;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
		this.h = Math.max(0, eval(h));
		this.CalcText();
		this.CalcOText();
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var str = this.func();
				if (str != this.SText) {
					this.Text = str;
					this.SText = str;
					this.Offset = 0;
					this.CalcText();
					this.CalcOText();
				}
			}
			this.reload = false;
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.h + 2 * marginy, 1, activecolour);
		var arr = gr.GdiDrawText(this.Text.substr(this.OBegin), this.Font, textcolour, this.x, this.y, this.w, this.h, DT_LEFT | DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
		if (this.hover) {
			var tfont = gdi.Font("Tahoma", 8);
			if (this.Offset > 0) {
				gr.FillSolidRect(this.x + this.w + marginx - 4, this.y, 7, 7, backcolour);
				gr.DrawRect(this.x + this.w + marginx - 4, this.y, 7, 7, 1, textcolour);
				gr.GdiDrawText("+", tfont, textcolour, this.x + this.w + marginx - 4, this.y - 1, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
			}
			if (this.Offset < this.TRows - this.Rows) {
				gr.FillSolidRect(this.x + this.w + marginx - 4, this.y + this.h - 8, 7, 7, backcolour);
				gr.DrawRect(this.x + this.w + marginx - 4, this.y + this.h - 8, 7, 7, 1, textcolour);
				gr.GdiDrawText("-", tfont, textcolour, this.x + this.w + marginx - 4, this.y + this.h - 9, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
			}
		}
	}

	this.on_mousemove = function (x, y) {
		var T = (this.x < x && x < this.x + this.w + marginx + 3 && this.y < y && y < this.y + this.h) ? true : false;
		var hov = (this.x < x && x < this.x + this.w && this.y < y && y < this.y + this.h) ? true : false;
		var U = (this.x + this.w < x && x < this.x + this.w + marginx + 3 && this.y < y && y < this.y + 8 && this.Offset > 0) ? true : false;
		var D = (this.x + this.w < x && x < this.x + this.w + marginx + 3 && this.y + this.h - 8 < y && y < this.y + this.h && this.Offset < this.TRows - this.Rows) ? true : false;
		if (T != this.hover || hov != this.hov || U != this.HovU || D != this.HovD) {
			var tcur = IDC_ARROW;
			if (hov) tcur = IDC_ARROW;
			if (U) tcur = IDC_HAND;
			if (D) tcur = IDC_HAND;
			window.SetCursor(tcur);
			this.hover = T;
			this.hov = hov;
			this.HovU = U;
			this.HovD = D;
			ommRepaint = true;
		}
		if (this.rbtnd) {
			var TOffset = this.SOffset + Math.ceil((this.ScrY - y) / this.THeight);
			TOffset = Math.min(TOffset, this.TRows - this.Rows);
			TOffset = Math.max(TOffset, 0);
			if (TOffset != this.Offset) {
				this.Offset = TOffset;
				ommRepaint = true;
				this.CalcOText();
			}
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hov) {
			if (!this.skip) this.active = true;
			omluRepaint = true;
		}
		if (this.HovU) {
			if (this.Scroll(1, false)) omluRepaint = true;
		}
		if (this.HovD) {
			if (this.Scroll(-1, false)) omluRepaint = true;
		}
		return this.hover;
	}

	this.on_mouse_rbtn_down = function (x, y) {
		this.ScrY = y;
		this.SOffset = this.Offset;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x, this.y, this.w, this.h);
			FormsAndButtons.Reload();
		}
		return true;
	}

	this.on_mouse_wheel = function (step) {
		if (this.hover) {
			if (this.Scroll(step, false)) omwRepaint = true;
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		this.CalcText();
		this.CalcOText();
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_PGUP:
				if (this.Scroll(1, true)) okdRepaint = true;
				break;
			case VK_PGDN:
				if (this.Scroll(-1, true)) okdRepaint = true;
				break;
			case VK_LEFT:
				FormsAndButtons.ActivatePrevious();
				okdRepaint = true;
				break;
			case VK_UP:
					if (!this.Scroll(1, false)) FormsAndButtons.ActivatePrevious();
				okdRepaint = true;
				break;
			case VK_RIGHT:
				FormsAndButtons.ActivateNext();
				okdRepaint = true;
				break;
			case VK_DOWN:
				if (this.Scroll(-1, false)) FormsAndButtons.ActivateNext();
				okdRepaint = true;
				break;
			}
		}
	}

	this.CalcText = function () {
		var arr = temp_gr.GdiDrawText(this.Text, this.Font, 0, 0, 0, this.w, 9999999, DT_VCENTER | DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
		this.TRows = (arr[3] - arr[1]) / this.THeight;
		this.Rows = Math.floor(this.h / this.THeight);
	}

	this.CalcOText = function () {
		if (this.Offset > 0) {
			var arr = temp_gr.GdiDrawText(this.Text, this.Font, 0, 0, 0, this.w, this.Offset * this.THeight, DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			this.OBegin = arr[4];
		} else this.OBegin = 0;
	}

	this.Scroll = function (n, end) {
		if (end) {
			if (n > 0) {
				var TOffset = 0;
			} else {
				var TOffset = this.TRows - this.Rows;
			}
		}
		else var TOffset = this.Offset - n;
		TOffset = Math.min(TOffset, this.TRows - this.Rows);
		TOffset = Math.max(TOffset, 0);
		if (TOffset != this.Offset) {
			this.Offset = TOffset;
			this.CalcOText();
			return true
		} else return false
	}
}

FormTextInput.prototype = new FormProto;
FormTextInput.prototype.constructor = FormTextInput;

function FormTextInput(x, y, w, skip, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.skip = skip;
	this.Text = "";
	this.SText = "";
	this.Font = font;
	this.func = func;
	this.rbfunc = rbfunc;
	this.Offset = 0;
	this.TLength = null;
	this.TWidth = null;
	this.THeight = null;
	this.CPos = 0;
	this.SPos = 0;
	this.SelBegin = 0;
	this.SelEnd = 0;
	this.Cx = 0;
	this.edit = false;
	this.insert = false;
	this.select = false;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var str = this.func();
				if (str != this.SText) {
					this.Text = str;
					this.SText = str;
					this.Offset = 0;
					this.edit = false;
					this.select = false;
					this.CalcText();
				}
			}
			this.reload = false;
		}
		if (this.edit) {
			gr.FillSolidRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, editbackcolour);
			this.ForceVisible();
			if (this.select) this.DrawSelection(gr);
			var colour = edittextcolour;
			var DT = DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT;
		} else {
			var colour = textcolour;
			var DT = DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS;
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
		gr.GdiDrawText(this.Text.substr(this.Offset), this.Font, colour, this.x, this.y, this.w, this.THeight, DT);
		if (this.edit) this.DrawCursor(gr);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		if (hov != this.hover) {
			if (hov) {
				if (this.w < this.TWidth) {
					tooltip.Text = this.Text;
					tooltip.Activate();
				}
			} else tooltip.Deactivate();
			window.SetCursor(hov ? IDC_IBEAM : IDC_ARROW);
			this.hover = hov;
		}
		if (this.lbtnd) {
			var tmp = this.CPos;
			this.GetCPos(x);
			if (this.Cpos != this.SPos) this.select = true;
			if (tmp != this.CPos) ommRepaint = true;
		}
	}

	this.on_mouse_lbtn_down = function (x, y) {
		if (!this.skip) this.active = true;
		this.edit = true;
		this.select = false;
		this.GetCPos(x);
		this.SPos = this.CPos;
		omldRepaint = true;
	}

	this.on_mouse_lbtn_up = function (x, y) {
		return true;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x, this.y, this.w, this.THeight);
			FormsAndButtons.Reload();
			return true;
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		if (!act && this.edit) {
			this.edit = false;
			if (this.Text != this.SText) if (this.func) this.func(this.Text);
		}
		this.CPos = 0;
		this.Select = false;
		this.Offset = 0;
		this.CalcText();
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_BACK:
				if (this.edit) {
					if (this.select) {
						this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText();
					} else {
						if (this.CPos > 0) {
							this.Text = this.Text.slice(0, this.CPos - 1) + this.Text.slice(this.CPos);
							this.CPos--;
							this.CalcText();
						}
					}
				}
				break;
			case VK_RETURN:
				if (this.edit) {
					this.edit = false;
					if (this.Text != this.SText) if (this.func) this.func(this.Text);
					okdRepaint = true;
				} else {
					this.edit = true;
					this.select = false;
					this.CPos = this.TLength;
				};
				break;
			case VK_ESCAPE:
				if (this.edit) {
					this.edit = false;
					if (this.Text != this.SText) if (this.func) this.func(this.Text);
					okdRepaint;
				}
				break;
			case VK_END:
				this.CPos = this.TLength;
				this.select = false;
				break;
			case VK_HOME:
				this.CPos = 0;
				this.Offset = 0;
				this.select = false;
				break;
			case VK_LEFT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelBegin;
						this.select = false;
					} else if (this.CPos > 0) this.CPos--;
				} else {
					FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_UP:
				if (this.edit) {
					this.select = false;
				} else {
					FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_RIGHT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelectEnd;
						this.select = false;
					} else if (this.CPos < this.TLength) this.CPos++;
				} else {
					FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_DOWN:
				if (this.edit) {
					this.select = false;
				} else {
					FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_INSERT:
				this.insert = !this.insert;
				break;
			case VK_DELETE:
				if (this.edit) {
					if (this.select) {
						this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText();
					} else {
						this.Text = this.Text.slice(0, this.CPos) + this.Text.slice(this.CPos + 1);
						this.CalcText();
					}
				}
				break;
			}
			if (this.edit) okdRepaint = true;
		}
		if (mask == KMask.shift && this.edit) {
			switch (vkey) {
			case VK_END:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = this.TLength;
				break;
			case VK_HOME:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = 0;
				this.Offset = 0;
				break;
			case VK_LEFT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos > 0) this.CPos--;
				break;
			case VK_RIGHT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos < this.TLength) this.CPos++;
				break;
			}
			if (this.edit) okdRepaint = true;
		}
	}

	this.on_char = function (code, mask) {
		if (code == 1 && this.edit && mask == KMask.ctrl) {
			// ctrl a
			this.SPos = 0;
			this.CPos = this.Text.length;
			this.select = true;
			ocRepaint = true;
		}
		if (code == 3 && this.edit && this.select && mask == KMask.ctrl) {
			// ctrl c
			FormsAndButtons.CopyBuffer = this.Text.substr(this.SelBegin, this.SelEnd - this.SelBegin);
		}
		if (code == 22 && this.edit && mask == KMask.ctrl) {
			// ctrl v
			if (FormsAndButtons.CopyBuffer.length > 0) {
				if (this.select) {
					var p1 = this.SelBegin;
					var p2 = this.SelEnd;
					this.CPos = this.SelBegin;
					this.select = false;
				} else {
					var p1 = this.CPos;
					var p2 = this.CPos;
					if (this.insert) p2 += FormsAndButtons.CopyBuffer.length;
					if (p2 > this.Text.length - 1) p2 = this.Text.length - 1;
				}
				this.Text = this.Text.slice(0, p1) + FormsAndButtons.CopyBuffer + this.Text.slice(p2);
				this.CPos += FormsAndButtons.CopyBuffer.length;
				this.select = false;
				this.CalcText();
				this.CalcOText();
				ocRepaint = true;
			}
		}
		if (code == 24 && this.edit && this.select && mask == KMask.ctrl) {
			// ctrl x
			FormsAndButtons.CopyBuffer = this.Text.substr(this.SelBegin, this.SelEnd - this.SelBegin);
			this.Text = this.Text.slice(0, this.SelBegin) + this.Text.slice(this.SelEnd);
			this.CPos = this.SelBegin;
			this.select = false;
			this.CalcText();
			this.CalcOText();
			ocRepaint = true;
		}
		if (code == 9) {
			// tab
			if (this.edit) {
				this.edit = false;
				if (this.Text != this.SText) if (this.func) this.func(this.Text);
			}
			if (mask == KMask.none) FormsAndButtons.ActivateNext();
			if (mask == KMask.shift) FormsAndButtons.ActivatePrevious();
			ocRepaint = true;
		}
		if (code == 26 && this.edit && mask == KMask.ctrl) {
			// ctrl z
			this.Text = this.SText;
			this.CPos = this.TLength;
			this.select = false;
			this.CalcText();
			ocRepaint = true;
		}
		if (code > 31 && this.edit) {
			if (this.select) {
				var p1 = this.SelBegin;
				var p2 = this.SelEnd;
				this.CPos = this.SelBegin;
				this.select = false;
			} else {
				var p1 = this.CPos;
				var p2 = this.CPos;
				if (this.insert) p2++;
			}
			this.Text = this.Text.slice(0, p1) + String.fromCharCode(code) + this.Text.slice(p2);
			this.CPos++;
			this.CalcText();
			ocRepaint = true;
		}
	}

	this.CalcText = function () {
		var arr = temp_gr.GdiDrawText(this.Text, this.Font, 0, 0, 0, 9999999, 9999999, DT_NOPREFIX | DT_CALCRECT).toArray();
		this.TLength = arr[4];
		this.TWidth = temp_gr.CalcTextWidth(this.Text, this.Font);
	}

	this.DrawSelection = function (gr) {
		var tmp = this.CPos > this.SPos;
		this.SelBegin = tmp ? this.SPos : this.CPos;
		this.SelEnd = tmp ? this.CPos : this.SPos;
		if (this.SelBegin < this.Offset) this.SelBegin = this.Offset;
		if (this.SelBegin != this.SelEnd) {
			var Sx1 = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.SelBegin - this.Offset), this.Font);
			var Sx2 = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.SelEnd - this.Offset), this.Font);
			gr.FillSolidRect(this.x + Sx1, this.y, Sx2 - Sx1, this.THeight, selectcolour);
		}
	}

	this.DrawCursor = function (gr) {
		if (this.insert) {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx + Math.round(0.3 * this.THeight);
			var y1 = this.y + this.THeight;
			var y2 = this.y + this.THeight;
			var lt = 2;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1 - 1;
			FormsAndButtons.crpw = x2 - x1 + 1;
			FormsAndButtons.crph = lt + 1;
		} else {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx;
			var y1 = this.y;
			var y2 = this.y + this.THeight;
			var lt = 1;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1;
			FormsAndButtons.crpw = lt;
			FormsAndButtons.crph = y2 - y1 + 1;
		}
		if (FormsAndButtons.cvisible) gr.DrawLine(x1, y1, x2, y2, lt, edittextcolour);
	}

	this.ForceVisible = function () {
		while (this.CPos < this.Offset) this.Offset--;
		this.Cx = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.CPos - this.Offset), this.Font);
		while (this.Cx > this.w) {
			this.Offset++;
			this.Cx = temp_gr.CalcTextWidth(this.Text.substr(this.Offset, this.CPos - this.Offset), this.Font);
		}
	}

	this.GetCPos = function (x) {
		x -= this.x;
		this.CPos = this.Offset;
		var trlen = temp_gr.CalcTextWidth(this.Text.substr(this.Offset), this.Font);
		if (x > trlen) this.CPos = this.TLength;
		else {
			var arr = temp_gr.GdiDrawText(this.Text.substr(this.Offset), this.Font, 0, 0, 0, this.w, this.THeight, DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			var tmp = x / trlen * arr[4];
			var td1 = x - temp_gr.CalcTextWidth(this.Text.substr(this.Offset, tmp), this.Font);
			var sign = td1 == 0 ? 1 : Math.abs(td1) / td1;
			var e = 0;
			do {
				tmp += sign;
				td2 = x - temp_gr.CalcTextWidth(this.Text.substr(this.Offset, tmp), this.Font);
				t = td1;
				td1 = td2;
				e++;
				if (e > 100) {
					fb.trace("FormTextInput.GetCPos");
					break
				}
			} while (Math.abs(td2) < Math.abs(t));
			this.CPos += Math.floor(tmp - sign);
		}
	}
}

FormTextLine.prototype = new FormProto;
FormTextLine.prototype.constructor = FormTextLine;

function FormTextLine(x, y, w, skip, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.skip = skip;
	this.Text = "";
	this.SText = "";
	this.Font = font;
	this.func = func;
	this.rbfunc = rbfunc;
	this.TLength = null;
	this.TWidth = null;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var str = this.func();
				if (str != this.SText) {
					this.Text = str;
					this.SText = str;
					this.CalcText();
				}
			}
			this.reload = false;
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
		gr.GdiDrawText(this.Text, this.Font, textcolour, this.x, this.y, this.w, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		if (hov != this.hover) {
			if (hov) {
				if (this.w < this.TWidth) {
					tooltip.Text = this.Text;
					tooltip.Activate();
				}
			} else tooltip.Deactivate();
			this.hover = hov;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			omluRepaint = true;
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y, this.x, this.y, this.w, this.THeight);
			FormsAndButtons.Reload();
			return true;
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		this.CalcText();
	}

	this.CalcText = function () {
		this.TWidth = temp_gr.CalcTextWidth(this.Text, this.Font);
	}
}

FormMVTextInput.prototype = new FormProto;
FormMVTextInput.prototype.constructor = FormMVTextInput;

function FormMVTextInput(x, y, w, skip, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.skip = skip;
	this.Values = [];
	this.SValues = [];
	this.PValues = [];
	this.PValuesP = [];
	this.PEvalues = [];
	this.LValues = [];
	this.LValuesP = [];
	this.ValuesStr = "";
	this.ValuesEStr = "";
	this.Font = font;
	this.func = func;
	this.rbfunc = rbfunc;
	this.Offset = 0;
	this.TLength = null;
	this.ETLenght = null;
	this.THeight = null;
	this.TWidth = null;
	this.CPos = 0;
	this.SPos = 0;
	this.SelBegin = 0;
	this.SelEnd = 0;
	this.Cx = 0;
	this.endspaces = "";
	this.edit = false;
	this.insert = false;
	this.select = false;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var arr = this.func();
				if (arr.toString() != this.SValues.toString()) {
					this.Values = arr.slice(0);
					this.SValues = arr.slice(0);
					this.Offset = 0;
					this.edit = false;
					this.select = false;
					this.CalcText();
				}
			}
			this.reload = false;
		}
		if (this.edit) {
			gr.FillSolidRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, editbackcolour);
			this.ForceVisible();
			if (this.select) this.DrawSelection(gr);
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
		var tmpx = 0;
		for (var i in this.Values) {
			if (this.edit) {
				if (i >= this.Offset) {
					if (i > this.Offset) {
						gr.GdiDrawText(this.ValuesEStr.substr(this.PEValues[i - 1] + this.LValuesP[i - 1], this.PEValues[i] - this.PEValues[i - 1] - this.LValuesP[i - 1]), this.Font, editsepcolour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT);
						tmpx += gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[i - 1] + this.LValuesP[i - 1], this.PEValues[i] - this.PEValues[i - 1] - this.LValuesP[i - 1]), this.Font);
					}
					gr.GdiDrawText(this.ValuesEStr.substr(this.PEValues[i], this.LValuesP[i]), this.Font, edittextcolour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT);
					tmpx += gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[i], this.LValuesP[i]), this.Font);
				}
			} else {
				if (i > 0) {
					gr.GdiDrawText(this.ValuesStr.substr(this.PValuesP[i - 1] + this.LValuesP[i - 1], this.PValuesP[i] - this.PValuesP[i - 1] - this.LValuesP[i - 1]), this.Font, sepcolour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
					tmpx += gr.CalcTextWidth(this.ValuesStr.substr(this.PValuesP[i - 1] + this.LValuesP[i - 1], this.PValuesP[i] - this.PValuesP[i - 1] - this.LValuesP[i - 1]), this.Font);
				}
				this.PValues[i] = this.x + tmpx;
				gr.GdiDrawText(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font, textcolour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
				this.LValues[i] = gr.CalcTextWidth(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font);
				tmpx += this.LValues[i];
			}
		}
		if (this.edit) this.DrawCursor(gr);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		if (hov != this.hover) {
			if (hov && !this.edit) {
				if (this.w < this.TWidth) {
					tooltip.Text = this.ValuesStr;
					tooltip.Activate();
				}
			} else tooltip.Deactivate();
			window.SetCursor(hov ? IDC_IBEAM : IDC_ARROW);
			this.hover = hov;
		}
		if (this.lbtnd) {
			var tmp = this.CPos;
			this.GetCPos(x);
			if (this.Cpos != this.SPos) this.select = true;
			if (tmp != this.CPos) ommRepaint = true;
		}
	}

	this.on_mouse_lbtn_down = function (x, y) {
		if (!this.skip) this.active = true;
		this.edit = true;
		this.select = false;
		this.GetCPos(x);
		this.SPos = this.CPos;
		omldRepaint = true;
	}

	this.on_mouse_lbtn_up = function (x, y) {
		return true;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.edit && this.hover && this.rbfunc) {
			this.rbfunc(x, y, -1, this.x, this.y, this.w, this.THeight);
			FormsAndButtons.Reload();
			return true;
		}
		if (!this.edit && this.hover) {
			var n = -1;
			for (var i in this.Values) {
				if (this.PValues[i] < x && x < (this.PValues[i] + this.LValues[i])) n = i;
			}
			if (this.rbfunc && n >= 0) {
				this.rbfunc(x, y, n, this.PValues[n], this.y, this.LValues[n], this.THeight);
				FormsAndButtons.Reload();
				return true;
			}
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		if (!act && this.edit) {
			this.edit = false;
			if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
		}
		this.CPos = 0;
		this.select = false;
		this.Offset = 0;
		this.CalcText();
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_BACK:
				if (this.edit) {
					if (this.select) {
						this.ValuesEStr = this.ValuesEStr.slice(0, this.SelBegin) + this.ValuesEStr.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText(true);
					} else {
						if (this.CPos > 0) {
							this.ValuesEStr = this.ValuesEStr.slice(0, this.CPos - 1) + this.ValuesEStr.slice(this.CPos);
							this.CPos--;
							this.CalcText(true);
						}
					}
				}
				break;
			case VK_RETURN:
				if (this.edit) {
					this.edit = false;
					if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
					okdRepaint = true;
				} else {
					this.edit = true;
					this.CPos = this.ETLength;
				};
				break;
			case VK_ESCAPE:
				if (this.edit) {
					this.edit = false;
					if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
					okdRepaint = true;
				}
				break;
			case VK_END:
				this.CPos = this.ETLength;
				this.select = false;
				break;
			case VK_HOME:
				this.CPos = 0;
				this.Offset = 0;
				this.select = false;
				break;
			case VK_LEFT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelBegin;
						this.select = false;
					} else if (this.CPos > 0) this.CPos--;
				} else {
					FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_UP:
				if (this.edit) {
					this.select = false;
				} else {
					FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_RIGHT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelEnd;
						this.select = false;
					} else if (this.CPos < this.ETLength) this.CPos++;
				} else {
					FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_DOWN:
				if (this.edit) {
					this.select = false;
				} else {
					FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_INSERT:
				this.insert = !this.insert;
				break;
			case VK_DELETE:
				if (this.edit) {
					if (this.select) {
						this.ValuesEStr = this.ValuesEStr.slice(0, this.SelBegin) + this.ValuesEStr.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText(true);
					} else {
						this.ValuesEStr = this.ValuesEStr.slice(0, this.CPos) + this.ValuesEStr.slice(this.CPos + 1);
						this.CalcText(true);
					}
				}
				break;
			}
			if (this.edit) okdRepaint = true;
		}
		if (mask == KMask.shift && this.edit) {
			switch (vkey) {
			case VK_END:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = this.ETLength;
				break;
			case VK_HOME:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = 0;
				this.Offset = 0;
				break;
			case VK_LEFT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos > 0) this.CPos--;
				break;
			case VK_RIGHT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos < this.ETLength) this.CPos++;
				break;
			}
			if (this.edit) okdRepaint = true;
		}
	}

	this.on_char = function (code, mask) {
		if (code == 1 && this.edit && mask == KMask.ctrl) {
			// ctrl a
			this.SPos = 0;
			this.CPos = this.ValuesEStr.length;
			this.select = true;
			ocRepaint = true;
		}
		if (code == 3 && this.edit && this.select && mask == KMask.ctrl) {
			// ctrl c
			FormsAndButtons.CopyBuffer = this.ValuesEStr.substr(this.SelBegin, this.SelEnd - this.SelBegin);
		}
		if (code == 22 && this.edit && mask == KMask.ctrl) {
			// ctrl v
			if (FormsAndButtons.CopyBuffer.length > 0) {
				if (this.select) {
					var p1 = this.SelBegin;
					var p2 = this.SelEnd;
					this.CPos = this.SelBegin;
					this.select = false;
				} else {
					var p1 = this.CPos;
					var p2 = this.CPos;
					if (this.insert) p2 += FormsAndButtons.CopyBuffer.length;
					if (p2 > this.ValuesEStr.length) p2 = this.ValuesEStr.length;
				}
				this.ValuesEStr = this.ValuesEStr.slice(0, p1) + FormsAndButtons.CopyBuffer + this.ValuesEStr.slice(p2);
				this.CPos += FormsAndButtons.CopyBuffer.length;
				this.select = false;
				this.CalcText(true);
				ocRepaint = true;
			}
		}
		if (code == 24 && this.edit && this.select && mask == KMask.ctrl) {
			// ctrl x
			FormsAndButtons.CopyBuffer = this.ValuesEStr.substr(this.SelBegin, this.SelEnd - this.SelBegin);
			this.ValuesEStr = this.ValuesEStr.slice(0, this.SelBegin) + this.ValuesEStr.slice(this.SelEnd);
			this.CPos = this.SelBegin;
			this.select = false;
			this.CalcText(true);
			ocRepaint = true;
		}
		if (code == 9) {
			// tab
			if (this.edit) {
				this.edit = false;
				if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
			}
			if (mask == KMask.none) FormsAndButtons.ActivateNext();
			if (mask == KMask.shift) FormsAndButtons.ActivatePrevious();
			ocRepaint = true;
		}
		if (code == 26 && this.edit && mask == KMask.ctrl) {
			// ctrl z
			this.Values = this.SValues.slice(0);
			this.CalcText();
			this.CPos = this.TLength;
			ocRepaint = true;
		}
		if (code > 31 && this.edit) {
			if (this.select) {
				var p1 = this.SelBegin;
				var p2 = this.SelEnd;
				this.CPos = this.SelBegin;
				this.select = false;
			} else {
				var p1 = this.CPos;
				var p2 = this.CPos;
				if (this.insert) p2++;
			}
			this.ValuesEStr = this.ValuesEStr.slice(0, p1) + String.fromCharCode(code) + this.ValuesEStr.slice(p2);
			this.CPos++;
			this.CalcText(true);
			ocRepaint = true;
		}
	}

	this.CalcText = function (edit) {
		if (arguments.length != 0) this.SplitInput();
		this.ValuesStr = this.Values.join(displaysep);
		var arr = temp_gr.GdiDrawText(this.ValuesStr, this.Font, 0, 0, 0, 9999999, 9999999, DT_NOPREFIX | DT_CALCRECT).toArray();
		this.TLength = arr[4];
		this.ValuesEStr = this.Values.join("; ") + this.endspaces;
		var arr = temp_gr.GdiDrawText(this.ValuesEStr, this.Font, 0, 0, 0, 9999999, 9999999, DT_NOPREFIX | DT_CALCRECT).toArray();
		this.ETLength = arr[4];
		this.PValues = [];
		this.PValuesP = [];
		this.PEValues = [];
		this.LValues = [];
		this.LValuesP = [];
		var tpos = 0;
		var etpos = 0;
		for (var i in this.Values) {
			this.PValuesP[i] = tpos;
			this.PEValues[i] = etpos;
			this.LValuesP[i] = this.Values[i].length;
			tpos += this.LValuesP[i] + displaysep.length;
			etpos += this.LValuesP[i] + 2
		}
		this.TWidth = temp_gr.CalcTextWidth(this.ValuesStr, this.Font);
	}

	this.DrawSelection = function (gr) {
		var tmp = this.CPos > this.SPos;
		this.SelBegin = tmp ? this.SPos : this.CPos;
		this.SelEnd = tmp ? this.CPos : this.SPos;
		if (this.SelBegin < this.Offset) this.SelBegin = this.Offset;
		if (this.SelBegin != this.SelEnd) {
			var Sx1 = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.Offset, this.SelBegin - this.Offset), this.Font);
			var Sx2 = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.Offset, this.SelEnd - this.Offset), this.Font);
			gr.FillSolidRect(this.x + Sx1, this.y, Sx2 - Sx1, this.THeight, selectcolour);
		}
	}

	this.DrawCursor = function (gr) {
		if (this.insert) {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx + Math.round(0.3 * this.THeight);
			var y1 = this.y + this.THeight;
			var y2 = this.y + this.THeight;
			var lt = 2;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1 - 1;
			FormsAndButtons.crpw = x2 - x1 + 1;
			FormsAndButtons.crph = lt + 1;
		} else {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx;
			var y1 = this.y;
			var y2 = this.y + this.THeight;
			var lt = 1;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1;
			FormsAndButtons.crpw = lt;
			FormsAndButtons.crph = y2 - y1 + 1;
		}
		if (FormsAndButtons.cvisible) gr.DrawLine(x1, y1, x2, y2, lt, edittextcolour);
	}

	this.ForceVisible = function () {
		while (this.CPos < this.PEValues[this.Offset]) this.Offset--;
		this.Cx = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[this.Offset], this.CPos - this.PEValues[this.Offset]), this.Font);
		while (this.Cx > this.w) {
			this.Offset++;
			this.Cx = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[this.Offset], this.CPos - this.PEValues[this.Offset]), this.Font);
		}
	}

	this.GetCPos = function (x) {
		x -= this.x;
		var tstr = this.edit ? this.ValuesEStr : this.ValuesStr;
		var pval = this.edit ? this.PEValues : this.PValuesP;
		this.CPos = pval[this.Offset];
		var trlen = temp_gr.CalcTextWidth(tstr.substr(pval[this.Offset]), this.Font);
		if (x > trlen) this.CPos = tstr.length;
		else {
			var arr = temp_gr.GdiDrawText(tstr.substr(pval[this.Offset]), this.Font, 0, 0, 0, this.w, this.THeight, DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			var tmp = x / trlen * arr[4];
			var td1 = x - temp_gr.CalcTextWidth(tstr.substr(pval[this.Offset], tmp), this.Font);
			var sign = td1 == 0 ? 1 : Math.abs(td1) / td1;
			var e = 0;
			do {
				tmp += sign;
				td2 = x - temp_gr.CalcTextWidth(tstr.substr(pval[this.Offset], tmp), this.Font);
				t = td1;
				td1 = td2;
				e++;
				if (e > 100) {
					fb.trace("ERROR FormMVTextInput.GetCPos");
					break
				}
			} while (Math.abs(td2) < Math.abs(t));
			this.CPos += Math.floor(tmp - sign);
		}
		if (!this.edit) {
			for (var i in this.Values) {
				if (this.CPos >= this.PValuesP[i] && this.CPos < this.PValuesP[i] + this.LValuesP[i] + displaysep.length) {
					this.CPos += this.PEValues[i] - this.PValuesP[i];
					break;
				}
			}
		}
	}

	this.SplitInput = function() {
		this.Values = [];
		var d = 0;
		var nb = 0;
		var ne = 0;
		var sp = true;
		var nsp = 0;
		for (var i = 0; i < this.ValuesEStr.length; i++) {
			if (this.ValuesEStr.substr(i, 1) == " " && sp) {
				nb++;
				nsp++;
				if (nsp > 1 && this.CPos > i) d--;
			} else sp = false;
			if (this.ValuesEStr.substr(i, 1) == ";") {
				ne = i;
				while (this.ValuesEStr.substr(ne - 1, 1) == " ") {
					if (ne == 0 || ne == nb) break;
					if (this.CPos >= ne) d--;
					ne--;
				}
				this.Values.push(this.ValuesEStr.substring(nb, ne));
				nb = i + 1;
				sp = true;
				nsp = 0;
				if (this.ValuesEStr.substr(i + 1, 1) != " " && this.CPos > i) d++;
			}
		}
		this.endspaces = "";
		while (this.ValuesEStr.substr(i - 1, 1) == " ") {
			if (i == 0 || i == nb) break;
			i--;
			this.endspaces += " ";
		}
		this.Values.push(this.ValuesEStr.substring(nb, i));
		this.CPos += d;
	}
}

FormMVSTextInput.prototype = new FormProto;
FormMVSTextInput.prototype.constructor = FormMVSTextInput;

function FormMVSTextInput(x, y, w, skip, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.skip = skip;
	this.Values = [];
	this.SValues = [];
	this.PValues = [];
	this.PValuesP = [];
	this.SValuesP = [];
	this.PValuesP = [];
	this.PEValues = [];
	this.LValues = [];
	this.LValuesP = [];
	this.LEValues = [];
	this.ValuesStr = "";
	this.ValuesEStr = "";
	this.Font = font;
	this.func = func;
	this.rbfunc = rbfunc;
	this.Offset = 0;
	this.TLength = null;
	this.ETLenght = null;
	this.THeight = null;
	this.TWidth = null;
	this.CPos = 0;
	this.SPos = 0;
	this.SelBegin = 0;
	this.SelEnd = 0;
	this.Cx = 0;
	this.endspaces = "";
	this.edit = false;
	this.insert = false;
	this.select = false;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var arr = this.func();
				if (arr.toString() != this.SValues.toString()) {
					this.Values = arr.slice(0);
					this.SValues = arr.slice(0);
					this.Offset = 0;
					this.edit = false;
					this.select = false;
					this.CalcText();
				}
			}
			this.reload = false;
		}
		if (this.edit) {
			gr.FillSolidRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, editbackcolour);
			if (this.select) this.DrawSelection(gr);
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
		if (this.edit) {
			this.ForceVisible();
			var tmpx = 0;
			for (var i in this.Values) {
				if (i >= this.Offset) {
					if (i > this.Offset) {
						gr.GdiDrawText(this.ValuesEStr.substr(this.PEValues[i - 1] + this.LEValues[i - 1], this.PEValues[i] - this.PEValues[i - 1] - this.LEValues[i - 1]), this.Font, editsepcolour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT);
						tmpx += gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[i - 1] + this.LEValues[i - 1], this.PEValues[i] - this.PEValues[i - 1] - this.LEValues[i - 1]), this.Font);
					}
					gr.GdiDrawText(this.ValuesEStr.substr(this.PEValues[i], this.LEValues[i]), this.Font, edittextcolour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT);
					tmpx += gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[i], this.LEValues[i]), this.Font);
				}
			}
			this.DrawCursor(gr);
		} else {
			var odd = true;
			tmpx = 0;
			for (var i in this.Values) {
				this.PValues[i] = this.x + tmpx;
				var colour = (odd) ? textcolour : sepcolour;
				gr.GdiDrawText(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font, colour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
				this.LValues[i] = gr.CalcTextWidth(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font);
				tmpx += this.LValues[i];
				odd = !odd;
			}
		}
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		if (hov != this.hover) {
			if (hov && !this.edit) {
				if (this.w < this.TWidth) {
					tooltip.Text = this.ValuesStr;
					tooltip.Activate();
				}
			} else tooltip.Deactivate();
			window.SetCursor(hov ? IDC_IBEAM : IDC_ARROW);
			this.hover = hov;
		}
		if (this.lbtnd) {
			var tmp = this.CPos;
			this.GetCPos(x);
			if (this.Cpos != this.SPos) this.select = true;
			if (tmp != this.CPos) ommRepaint = true;
		}
	}

	this.on_mouse_lbtn_down = function (x, y) {
		if (!this.skip) this.active = true;
		this.edit = true;
		this.select = false;
		this.GetCPos(x);
		this.SPos = this.CPos;
		omldRepaint = true;
	}

	this.on_mouse_lbtn_up = function (x, y) {
		return true;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.edit && this.hover && this.rbfunc) {
			this.rbfunc(x, y, -1, this.x, this.y, this.w, this.THeight);
			FormsAndButtons.Reload();
			return true;
		}
		if (!this.edit && this.hover) {
			var n = -1;
			for (var i in this.Values) {
				if (this.PValues[i] < x && x < (this.PValues[i] + this.LValues[i])) n = i;
			}
			if (this.rbfunc && n >= 0) {
				this.rbfunc(x, y, n, this.PValues[n], this.y, this.LValues[n], this.THeight);
				FormsAndButtons.Reload();
				return true;
			}
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		if (!act && this.edit) {
			this.edit = false;
			if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
		}
		this.CPos = 0;
		this.select = false;
		this.Offset = 0;
		this.CalcText();
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_BACK:
				if (this.edit) {
					if (this.select) {
						this.ValuesEStr = this.ValuesEStr.slice(0, this.SelBegin) + this.ValuesEStr.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText(true);
					} else {
						if (this.CPos > 0) {
							this.ValuesEStr = this.ValuesEStr.slice(0, this.CPos - 1) + this.ValuesEStr.slice(this.CPos);
							this.CPos--;
							this.CalcText(true);
						}
					}
				}
				break;
			case VK_RETURN:
				if (this.edit) {
					this.edit = false;
					if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
					okdRepaint = true;
				} else {
					this.edit = true;
					this.CPos = this.ETLength;
				};
				break;
			case VK_ESCAPE:
				if (this.edit) {
					this.edit = false;
					if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
					okdRepaint = true;
				}
				break;
			case VK_END:
				this.CPos = this.ETLength;
				this.select = false;
				break;
			case VK_HOME:
				this.CPos = 0;
				this.Offset = 0;
				this.select = false;
				break;
			case VK_LEFT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelBegin;
						this.select = false;
					} else if (this.CPos > 0) this.CPos--;
				} else {
					FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_UP:
				if (this.edit) {
					this.select = false;
				} else {
					FormsAndButtons.ActivatePrevious();
					okdRepaint = true;
				}
				break;
			case VK_RIGHT:
				if (this.edit) {
					if (this.select) {
						this.CPos = this.SelEnd;
						this.select = false;
					} else if (this.CPos < this.ETLength) this.CPos++;
				} else {
					FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_DOWN:
				if (this.edit) {
					this.select = false;
				} else {
					FormsAndButtons.ActivateNext();
					okdRepaint = true;
				}
				break;
			case VK_INSERT:
				this.insert = !this.insert;
				break;
			case VK_DELETE:
				if (this.edit) {
					if (this.select) {
						this.ValuesEStr = this.ValuesEStr.slice(0, this.SelBegin) + this.ValuesEStr.slice(this.SelEnd);
						this.CPos = this.SelBegin;
						this.select = false;
						this.CalcText(true);
					} else {
						this.ValuesEStr = this.ValuesEStr.slice(0, this.CPos) + this.ValuesEStr.slice(this.CPos + 1);
						this.CalcText(true);
					}
				}
				break;
			}
			if (this.edit) okdRepaint = true;
		}
		if (mask == KMask.shift && this.edit) {
			switch (vkey) {
			case VK_END:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = this.ETLength;
				break;
			case VK_HOME:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				this.CPos = 0;
				this.Offset = 0;
				break;
			case VK_LEFT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos > 0) this.CPos--;
				break;
			case VK_RIGHT:
				if (!this.select) {
					this.SPos = this.CPos;
					this.select = true;
				}
				if (this.CPos < this.ETLength) this.CPos++;
				break;
			}
			if (this.edit) okdRepaint = true;
		}
	}

	this.on_char = function (code, mask) {
		if (code == 1 && this.edit && mask == KMask.ctrl) {
			// ctrl a
			this.SPos = 0;
			this.CPos = this.ValuesEStr.length;
			this.select = true;
			ocRepaint = true;
		}
		if (code == 3 && this.edit && this.select && mask == KMask.ctrl) {
			// ctrl c
			FormsAndButtons.CopyBuffer = this.ValuesEStr.substr(this.SelBegin, this.SelEnd - this.SelBegin);
		}
		if (code == 22 && this.edit && mask == KMask.ctrl) {
			// ctrl v
			if (FormsAndButtons.CopyBuffer.length > 0) {
				if (this.select) {
					var p1 = this.SelBegin;
					var p2 = this.SelEnd;
					this.CPos = this.SelBegin;
					this.select = false;
				} else {
					var p1 = this.CPos;
					var p2 = this.CPos;
					if (this.insert) p2 += FormsAndButtons.CopyBuffer.length;
					if (p2 > this.ValuesEStr.length) p2 = this.ValuesEStr.length;
				}
				this.ValuesEStr = this.ValuesEStr.slice(0, p1) + FormsAndButtons.CopyBuffer + this.ValuesEStr.slice(p2);
				this.CPos += FormsAndButtons.CopyBuffer.length;
				this.select = false;
				this.CalcText(true);
				ocRepaint = true;
			}
		}
		if (code == 24 && this.edit && this.select && mask == KMask.ctrl) {
			// ctrl x
			FormsAndButtons.CopyBuffer = this.ValuesEStr.substr(this.SelBegin, this.SelEnd - this.SelBegin);
			this.ValuesEStr = this.ValuesEStr.slice(0, this.SelBegin) + this.ValuesEStr.slice(this.SelEnd);
			this.CPos = this.SelBegin;
			this.select = false;
			this.CalcText(true);
			ocRepaint = true;
		}
		if (code == 9) {
			// tab
			if (this.edit) {
				this.edit = false;
				if (this.Values.toString() != this.SValues.toString()) if (this.func) this.func(this.Values);
			}
			if (mask == KMask.none) FormsAndButtons.ActivateNext();
			if (mask == KMask.shift) FormsAndButtons.ActivatePrevious();
			ocRepaint = true;
		}
		if (code == 26 && this.edit && mask == KMask.ctrl) {
			// ctrl z
			this.Values = this.SValues.slice(0);
			this.CalcText();
			this.CPos = this.TLength;
			ocRepaint = true;
		}
		if (code > 31 && this.edit) {
			if (this.select) {
				var p1 = this.SelBegin;
				var p2 = this.SelEnd;
				this.CPos = this.SelBegin;
				this.select = false;
			} else {
				var p1 = this.CPos;
				var p2 = this.CPos;
				if (this.insert) p2++;
			}
			this.ValuesEStr = this.ValuesEStr.slice(0, p1) + String.fromCharCode(code) + this.ValuesEStr.slice(p2);
			this.CPos++;
			this.CalcText(true);
			ocRepaint = true;
		}
	}

	this.DrawSelection = function (gr) {
		var tmp = this.CPos > this.SPos;
		this.SelBegin = tmp ? this.SPos : this.CPos;
		this.SelEnd = tmp ? this.CPos : this.SPos;
		if (this.SelBegin < this.Offset) this.SelBegin = this.Offset;
		if (this.SelBegin != this.SelEnd) {
			var Sx1 = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.Offset, this.SelBegin - this.Offset), this.Font);
			var Sx2 = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.Offset, this.SelEnd - this.Offset), this.Font);
			gr.FillSolidRect(this.x + Sx1, this.y, Sx2 - Sx1, this.THeight, selectcolour);
		}
	}

	this.CalcText = function (edit) {
		if (arguments.length != 0) this.SplitInput();
		this.ValuesStr = "";
		this.ValuesEStr = this.Values.join("; ") + this.endspaces;
		this.PValues = [];
		this.PValuesP = [];
		this.PEValues = [];
		this.LValues = [];
		this.LValuesP = [];
		this.LEValues = [];
		var tpos = 0;
		var etpos = 0;
		var odd = true;
		for (var i in this.Values) {
			this.PValuesP[i] = tpos;
			this.PEValues[i] = etpos;
			this.LEValues[i] = this.Values[i].length;
			if (odd) {
				this.LValuesP[i] = this.Values[i].length;
				this.ValuesStr += this.Values[i];
				tpos += this.LValuesP[i];
			} else {
				var stmp = this.Values[i];
				if (stmp == "") stmp = ";";
				if (stmp.charAt(0) != ";" && stmp.charAt(0) != ":" && stmp.charAt(0) != "," && stmp.charAt(0) != ".") stmp = " " + stmp;
				stmp += " ";
				this.LValuesP[i] = stmp.length;
				this.ValuesStr += stmp;
				tpos += stmp.length;
			}
			etpos += this.LEValues[i] + 2;
			odd = !odd;
		}
		var arr = temp_gr.GdiDrawText(this.ValuesStr, this.Font, 0, 0, 0, 9999999, 9999999, DT_NOPREFIX | DT_CALCRECT).toArray();
		this.TLength = arr[4];
		var arr = temp_gr.GdiDrawText(this.ValuesEStr, this.Font, 0, 0, 0, 9999999, 9999999, DT_NOPREFIX | DT_CALCRECT).toArray();
		this.ETLength = arr[4];
		this.TWidth = temp_gr.CalcTextWidth(this.ValuesStr, this.Font);
	}

	this.DrawCursor = function (gr) {
		if (this.insert) {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx + Math.round(0.3 * this.THeight);
			var y1 = this.y + this.THeight;
			var y2 = this.y + this.THeight;
			var lt = 2;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1 - 1;
			FormsAndButtons.crpw = x2 - x1 + 1;
			FormsAndButtons.crph = lt + 1;
		} else {
			var x1 = this.x + this.Cx;
			var x2 = this.x + this.Cx;
			var y1 = this.y;
			var y2 = this.y + this.THeight;
			var lt = 1;
			FormsAndButtons.crpx = x1;
			FormsAndButtons.crpy = y1;
			FormsAndButtons.crpw = lt;
			FormsAndButtons.crph = y2 - y1 + 1;
		}
		if (FormsAndButtons.cvisible) gr.DrawLine(x1, y1, x2, y2, lt, edittextcolour);
	}

	this.ForceVisible = function () {
		while (this.CPos < this.PEValues[this.Offset]) this.Offset--;
		this.Cx = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[this.Offset], this.CPos - this.PEValues[this.Offset]), this.Font);
		while (this.Cx > this.w) {
			this.Offset++;
			this.Cx = temp_gr.CalcTextWidth(this.ValuesEStr.substr(this.PEValues[this.Offset], this.CPos - this.PEValues[this.Offset]), this.Font);
		}
	}

	this.GetCPos = function (x) {
		x -= this.x;
		var tstr = this.edit ? this.ValuesEStr : this.ValuesStr;
		var pval = this.edit ? this.PEValues : this.PValuesP;
		this.CPos = pval[this.Offset];
		var trlen = temp_gr.CalcTextWidth(tstr.substr(pval[this.Offset]), this.Font);
		if (x > trlen) this.CPos = this.ValuesEStr.length;
		else {
			var arr = temp_gr.GdiDrawText(tstr.substr(pval[this.Offset]), this.Font, 0, 0, 0, this.w, this.THeight, DT_WORDBREAK | DT_NOPREFIX | DT_EDITCONTROL | DT_CALCRECT).toArray();
			var tmp = x / trlen * arr[4];
			var td1 = x - temp_gr.CalcTextWidth(tstr.substr(pval[this.Offset], tmp), this.Font);
			var sign = td1 == 0 ? 1 : Math.abs(td1) / td1;
			var e = 0;
			do {
				tmp += sign;
				td2 = x - temp_gr.CalcTextWidth(tstr.substr(pval[this.Offset], tmp), this.Font);
				t = td1;
				td1 = td2;
				e++;
				if (e > 100) {
					fb.trace("ERROR FormMVSTextInput.GetCPos");
					break
				}
			} while (Math.abs(td2) < Math.abs(t));
			this.CPos += Math.floor(tmp - sign);
		}
		if (!this.edit) {
			for (var i in this.Values) {
				if (this.CPos >= this.PValuesP[i] && this.CPos < this.PValuesP[i] + this.LValuesP[i]) {
					this.CPos += this.PEValues[i] - this.PValuesP[i];
					break;
				}
			}
		}
	}

	this.SplitInput = function() {
		this.Values = [];
		var d = 0;
		var nb = 0;
		var ne = 0;
		var sp = true;
		var nsp = 0;
		for (var i = 0; i < this.ValuesEStr.length; i++) {
			if (this.ValuesEStr.substr(i, 1) == " " && sp) {
				nb++;
				nsp++;
				if (nsp > 1 && this.CPos > i) d--;
			} else sp = false;
			if (this.ValuesEStr.substr(i, 1) == ";") {
				ne = i;
				while (this.ValuesEStr.substr(ne - 1, 1) == " ") {
					if (ne == 0 || ne == nb) break;
					if (this.CPos >= ne) d--;
					ne--;
				}
				this.Values.push(this.ValuesEStr.substring(nb, ne));
				nb = i + 1;
				sp = true;
				nsp = 0;
				if (this.ValuesEStr.substr(i + 1, 1) != " " && this.CPos > i) d++;
			}
		}
		this.endspaces = "";
		while (this.ValuesEStr.substr(i - 1, 1) == " ") {
			if (i == 0 || i == nb) break;
			i--;
			this.endspaces += " ";
		}
		this.Values.push(this.ValuesEStr.substring(nb, i));
		this.CPos += d;
	}
}

FormMVTextLine.prototype = new FormProto;
FormMVTextLine.prototype.constructor = FormMVTextLine;

function FormMVTextLine(x, y, w, skip, font, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.skip = skip;
	this.Values = [];
	this.SValues = [];
	this.PValues = [];
	this.LValues = [];
	this.PValuesP = [];
	this.LValuesP = [];
	this.ValuesStr = "";
	this.Font = font;
	this.func = func;
	this.rbfunc = rbfunc;
	this.THeight = null;
	this.TWidth = null;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var arr = this.func();
				if (arr.toString() != this.SValues.toString()) {
					this.Values = arr.slice(0);
					this.SValues = arr.slice(0);
					this.CalcText();
				}
			}
			this.reload = false;
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
		var odd = true;
		var tmpx = 0;
		for (var i in this.Values) {
			this.PValues[i] = this.x + tmpx;
			var colour = (odd) ? textcolour : sepcolour;
			gr.GdiDrawText(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font, colour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
			this.LValues[i] = gr.CalcTextWidth(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font);
			tmpx += this.LValues[i];
			odd = !odd;
		}
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		if (hov != this.hover) {
			if (hov) {
				if (this.w < this.TWidth) {
					tooltip.Text = this.ValuesStr;
					tooltip.Activate();
				}
			} else tooltip.Deactivate();
			this.hover = hov;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			omluRepaint = true;
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover) {
			var n = -1;
			var m = -1;
			var odd = true;
			var j = 0;
			for (var i in this.Values) {
				if (odd) {
					if (this.PValues[i] < x && x < (this.PValues[i] + this.LValues[i])) {
						n = j;
						m = i;
					}
					j++;
				}
				odd = !odd;
			}
			if (this.rbfunc && n >= 0) {
				this.rbfunc(x, y, n, this.PValues[m], this.y, this.LValues[m], this.THeight);
				FormsAndButtons.Reload();
				return true;
			}
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		this.CalcText();
	}

	this.CalcText = function () {
		this.ValuesStr = "";
		this.PValues = [];
		this.LValues = [];
		this.PValuesP = [];
		this.LValuesP = [];
		var tpos = 0;
		var odd = true;
		for (var i in this.Values) {
			this.PValuesP[i] = tpos;
			if (odd) {
				this.LValuesP[i] = this.Values[i].length;
				this.ValuesStr += this.Values[i];
				tpos += this.LValuesP[i];
			} else {
				var stmp = this.Values[i];
				if (stmp == "") stmp = ";";
				if (stmp.charAt(0) != ";" && stmp.charAt(0) != ":" && stmp.charAt(0) != "," && stmp.charAt(0) != ".") stmp = " " + stmp;
				stmp += " ";
				this.LValuesP[i] = stmp.length;
				this.ValuesStr += stmp;
				tpos += stmp.length;
			}
			odd = !odd;
		}
		this.TWidth = temp_gr.CalcTextWidth(this.ValuesStr, this.Font);
	}
}

FormKeyWords.prototype = new FormProto;
FormKeyWords.prototype.constructor = FormKeyWords;

function FormKeyWords(x, y, w, skip, font, listfunc, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = Math.max(0, eval(w));
	this.skip = skip;
	this.Values = [];
	this.SValues = [];
	this.PValues = [];
	this.LValues = [];
	this.PValuesP = [];
	this.LValuesP = [];
	this.ValuesStr = "";
	this.Font = font;
	this.listfunc = listfunc;
	this.list = [];
	this.func = func;
	this.rbfunc = rbfunc;
	this.Offset = 0;
	this.TLength = null;
	this.THeight = null;
	this.TWidth = null;
	this.ButX = [];
	this.ButY = 0;
	this.ButY = 0;
	this.HovE = null;
	this.HovB = null;
	this.ActE = 0;
	this.ActB = 0;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.w = Math.max(0, eval(w));
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.listfunc) this.list = this.listfunc();
			if (this.func) {
				var arr = this.func();
				if (arr.toString() != this.SValues.toString()) {
					this.Values = arr.slice(0);
					this.SValues = arr.slice(0);
					this.Offset = 0;
					this.ActE = 0;
					this.ActB = 0;
					this.CalcText();
				}
			}
			this.reload = false;
		}
		while (this.ActE < this.Offset) this.Offset--;
		while (gr.CalcTextWidth(this.ValuesStr.substr(this.PValuesP[this.Offset], this.PValuesP[this.ActE] + this.LValuesP[this.ActE]), this.Font) > this.w) this.Offset++;
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
		var tmpx = 0;
		this.ButX = [];
		this.ButY = this.y - marginy - 3;
		var tfont = gdi.Font("Tahoma", 8);
		for (var i in this.Values) {
			this.PValues[i] = 0;
			this.LValues[i] = 0;
			if (i >= this.Offset) {
				if (i > this.Offset) {
					gr.GdiDrawText(this.ValuesStr.substr(this.PValuesP[i - 1] + this.LValuesP[i - 1], this.PValuesP[i] - this.PValuesP[i - 1] - this.LValuesP[i - 1]), this.Font, sepcolour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
					tmpx += gr.CalcTextWidth(this.ValuesStr.substr(this.PValuesP[i - 1] + this.LValuesP[i - 1], this.PValuesP[i] - this.PValuesP[i - 1] - this.LValuesP[i - 1]), this.Font);
				}
				var colour = errorcolour;
				for (var j in this.list) {
					if (this.Values[i] == this.list[j]) {
						colour = textcolour;
						break;
					}
				}
				this.PValues[i] = this.x + tmpx;
				gr.GdiDrawText(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font, colour, this.x + tmpx, this.y, this.w - tmpx, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
				this.LValues[i] = gr.CalcTextWidth(this.ValuesStr.substr(this.PValuesP[i], this.LValuesP[i]), this.Font);
				if (this.LValues[i] >= 20) {
					this.ButX[i * 4] = this.x + tmpx - 4;
					this.ButX[i * 4 + 1] = this.x + tmpx + 3;
				} else {
					this.ButX[i * 4] = null;
					this.ButX[i * 4 + 1] = null;
				}
				if (this.LValues[i] >= 6) {
					this.ButX[i * 4 + 2] = this.x + tmpx + this.LValues[i] - 10;
					this.ButX[i * 4 + 3] = this.x + tmpx + this.LValues[i] - 3;
				} else {
					this.ButX[i * 4 + 2] = Math.round(this.x + tmpx + this.LValues[i] / 2 - 7);
					this.ButX[i * 4 + 3] = Math.round(this.x + tmpx + this.LValues[i] / 2);
				}
				if ((this.active && this.ActE == i) || this.HovE == i) {
					if (this.ActE == i && this.ActB < 2 && this.LValues[i] < 20) this.ActB = 2;
					for (var j = 0; j < 4; j++) {
						var colour = ((this.active && this.ActE == i) && this.ActB == j) ? activecolour : backcolour;
						if (this.ButX[i * 4 + j]) gr.FillSolidRect(this.ButX[i * 4 + j], this.ButY, 7, 7, colour);
					}
					if (this.ButX[i * 4]) gr.GdiDrawText("<", tfont, textcolour, this.ButX[i * 4], this.ButY - 1, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
					if (this.ButX[i * 4 + 1]) gr.GdiDrawText(">", tfont, textcolour, this.ButX[i * 4 + 1], this.ButY - 1, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
					gr.GdiDrawText("+", tfont, textcolour, this.ButX[i * 4 + 2], this.ButY - 1, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
					gr.GdiDrawText("-", tfont, textcolour, this.ButX[i * 4 + 3], this.ButY - 1, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
				}
				tmpx += this.LValues[i];
			}
		}
		if (this.Values.length == 0) {
			if (this.ActB < 2) this.ActB = 2;
			this.ButX[0] = null;
			this.ButX[1] = null;
			this.ButX[2] = this.x - 4;
			this.ButX[3] = null;
			if (this.active || this.hover) {
				var colour = this.active ? activecolour : backcolour;
				gr.FillSolidRect(this.ButX[2], this.ButY, 7, 7, colour);
				gr.GdiDrawText("+", tfont, textcolour, this.ButX[2], this.ButY - 1, 8, 8, DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_CALCRECT);
			}
		}
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x - 7 < x && x < (this.x + this.w + 4) && this.y - marginy - 4 <= y && y < (this.y + this.THeight)) ? true : false;
		var HE = this.HovE;
		var HB = this.HovB;
		this.HovE = -1;
		this.HovB = -1;
		if (hov) {
			for (var i in this.Values) {
				var tmpx = this.ButX[i * 4] ? this.ButX[i * 4] : this.ButX[i * 4 + 2]
				if (i >= this.Offset && this.ButX[i * 4 + 3] < this.x + this.w - 4 && x >= tmpx && x < this.ButX[i * 4 + 3] + 8) {
					this.HovE = i;
					if (this.ButX[i * 4]) if (x < this.ButX[i * 4] + 8 && y < this.ButY + 8) this.HovB = 0;
					if (this.ButX[i * 4 + 1]) if (x >= this.ButX[i * 4 + 1] && x < this.ButX[i * 4 + 1] + 8 && y < this.ButY + 8) this.HovB = 1;
					if (x >= this.ButX[i * 4 + 2] && x < this.ButX[i * 4 + 2] + 8 && y < this.ButY + 8) this.HovB = 2;
					if (x >= this.ButX[i * 4 + 3] && x < this.ButX[i * 4 + 3] + 8 && y < this.ButY + 8) this.HovB = 3;
				}
			}
			if (this.Values.length == 0) {
				if (x >= this.ButX[2] && x < this.ButX[2] + 8 && y < this.ButY + 8) this.HovB = 2;
			}
		}
		if (hov != this.hover) {
			if (hov) {
				if (this.w < this.TWidth) {
					tooltip.Text = this.ValuesStr;
					tooltip.Activate();
				}
			} else tooltip.Deactivate();
		}
		if (this.hover != hov || this.HovE != HE || this.HovB != HB) {
			window.SetCursor(this.HovB >= 0 ? IDC_HAND : IDC_ARROW);
			this.hover = hov;
			ommRepaint = true;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			if (this.Values.length == 0) {
				if (this.HovB == 2) this.AddItem (-1, x, y);
			} else {
				if (this.HovE >= 0) {
					this.ActE = this.HovE;
					switch (this.HovB) {
					case 0:
						this.MoveBackward(this.HovE);
						break;
					case 1:
						this.MoveForward(this.HovE);
						break;
					case 2:
						this.AddItem(this.HovE, x, y);
						break;
					case 3:
						this.DeleteItem(this.HovE);
						break;
					}
				}
			}
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover) {
		var n = this.HovE;
			if (this.rbfunc) {
				this.rbfunc(x, y, n, this.PValues[n], this.y, this.LValues[n], this.THeight);
				FormsAndButtons.Reload();
				return true;
			}
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		this.ActE = 0;
		this.ActB = 0;
		this.Offset = 0;
		this.CalcText();
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_END:
				this.ActE = this.Values.length - 1;
				this.ActB = 0;
				break;
			case VK_HOME:
				this.ActE = 0;
				this.ActB = 0;
				break;
			case VK_LEFT:
				if (mask == KMask.ctrl) this.MoveBackward(this.ActE);
				else {
					this.ActB--;
					if (this.ActB < 0 || !this.ButX[this.ActE * 4 + this.ActB]) {
						this.ActB = 3;
						this.ActE--;
						if (this.ActE < 0) FormsAndButtons.ActivatePrevious();
					}
				}
				break;
			case VK_UP:
				FormsAndButtons.ActivatePrevious();
				break;
			case VK_RIGHT:
				if (mask == KMask.ctrl) this.MoveForward(this.ActE);
				else {
					this.ActB++;
					if (this.ActB > 3 || (this.Values.length == 0 && this.ActB == 3)) {
						this.ActB = 0;
						this.ActE++;
						if (this.ActE >= this.Values.length) FormsAndButtons.ActivateNext();
					}
				}
				break;
			case VK_DOWN:
				FormsAndButtons.ActivateNext();
				break;
			case VK_INSERT:
				this.AddItem(this.ActE);
				break;
			case VK_DELETE:
				this.DeleteItem(this.ActE);
				break;
			}
			okdRepaint = true;
		}
	}

	this.on_char = function (code, mask) {
		if (code == 9) {
			// tab
			if (mask == KMask.none) FormsAndButtons.ActivateNext();
			if (mask == KMask.shift) FormsAndButtons.ActivatePrevious();
			ocRepaint = true;
		}
		if (code == 13 && mask == KMask.none) {
			switch (this.ActB) {
			case 0:
				this.MoveBackward(this.ActE);
				break;
			case 1:
				this.MoveForward(this.ActE);
				break;
			case 2:
				this.AddItem(this.ActE, this.ButX[this.ActE * 4 + 2] + 3, this.ButY + 3);
				break;
			case 3:
				this.DeleteItem(this.ActE);
				break;
			}
			ocRepaint = true;
		}
	}

	this.CalcText = function () {
		this.ValuesStr = this.Values.join(displaysep);
		var arr = temp_gr.GdiDrawText(this.ValuesStr, this.Font, 0, 0, 0, 9999999, 9999999, DT_NOPREFIX | DT_CALCRECT).toArray();
		this.TLength = arr[4];
		this.PValues = [];
		this.LValues = [];
		this.PValuesP = [];
		this.LValuesP = [];
		var tpos = 0;
		for (var i in this.Values) {
			this.PValuesP[i] = tpos;
			this.LValuesP[i] = this.Values[i].length;
			tpos += this.LValuesP[i] + displaysep.length;
		}
		this.TWidth = temp_gr.CalcTextWidth(this.ValuesStr, this.Font);
	}

	this.MoveBackward = function (item) {
		if (item > 0) {
			var tmp = this.Values[item - 1]
			this.Values[item - 1] = this.Values[item]
			this.Values[item] = tmp;
			if (this.func) {
				this.func(this.Values);
				FormsAndButtons.Reload();
			}
			this.ActE = item - 1;
			this.SValues = this.Values.slice(0);
			this.CalcText();
		}
	}

	this.MoveForward = function (item) {
		if (item < this.Values.length - 1) {
			var tmp = this.Values[parseInt(item) + 1]
			this.Values[parseInt(item) + 1] = this.Values[item]
			this.Values[item] = tmp;
			this.ActE = item + 1;
			this.SValues = this.Values.slice(0);
			if (this.func) {
				this.func(this.Values);
				FormsAndButtons.Reload();
			}
			this.CalcText();
		}
	}

	this.AddItem = function (item, x, y) {
		var popupmenu = window.CreatePopupMenu();
		for (var i in this.list) {
			popupmenu.AppendMenuItem(0, parseInt(i) + 1, this.list[i]);
		}
		var menuret = popupmenu.TrackPopupMenu(x, y);
		popupmenu.Dispose();
		if (menuret != 0) {
			this.Values.splice(parseInt(item) + 1, 0, this.list[menuret - 1]);
			if (this.func) {
				this.func(this.Values);
				FormsAndButtons.Reload();
			}
			if (this.Values.length > 1) this.ActE = item + 1;
			this.ActB = 0;
			this.SValues = this.Values.slice(0);
			this.CalcText();
		}
	}

	this.DeleteItem = function (item) {
		this.Values.splice(item, 1);
		if (this.func) {
			this.func(this.Values);
			FormsAndButtons.Reload();
		}
		if (this.ActE == item && item > 0) this.ActE--;
		this.SValues = this.Values.slice(0);
		this.CalcText();
	}
}

FormKeyWordsH.prototype = new FormProto;
FormKeyWordsH.prototype.constructor = FormKeyWordsH;

function FormKeyWordsH(x, y, w, skip, font, listfunc, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.mw = Math.max(0, eval(w));
	this.w = null;
	this.skip = skip;
	this.Values = [];
	this.SValues = [];
	this.PValues = [];
	this.LValues = [];
	this.Font = font;
	this.listfunc = listfunc;
	this.list = [];
	this.func = func;
	this.rbfunc = rbfunc;
	this.THeight = null;
	this.HovE = null;
	this.ActE = 0;
	this.tooltip = false;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];
	for (var i in this.list) this.Values.push(false);

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.mw = Math.max(0, eval(w));
	}
	this.draw = function (gr) {
		if (this.reload) {
			if (this.listfunc) this.list = this.listfunc();
			if (this.func) {
				var arr = this.func();
				if (arr.toString() != this.SValues.toString()) {
					this.SValues = arr.slice(0);
					this.CalcValues();
				}
			}
			this.reload = false;
		}
		var tmpx = 0;
		this.PValues = [];
		this.LValues = [];
		for (var i in this.list) {
			if (i > 0) tmpx += marginx;
			this.PValues[i] = this.x + tmpx
			this.LValues[i] = gr.CalcTextWidth(this.list[i], this.Font);
			tmpx += this.LValues[i];
		}
		this.w = tmpx;
		if (this.w > this.mw && this.mw > 0) {
			var d = (this.w - this.mw) / this.list.length;
			this.w = this.mw;
			this.tooltip = true;
		} else {
			var d = 0;
			this.tooltip = false;
		}
		for (var i in this.list) {
			this.PValues[i] = this.PValues[i] - i * d;
			this.LValues[i] = this.LValues[i] - d;
			var colour = this.Values[i] ? textcolour : uncheckedcolour;
			gr.GdiDrawText(this.list[i], this.Font, colour, this.PValues[i], this.y, this.LValues[i], this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
			if (this.active && this.ActE == i) gr.DrawRect(this.PValues[i] - marginx / 2, this.y - marginy / 2, this.LValues[i] + marginx, this.THeight + marginy, 1, activecolour);
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		var HE = this.HovE;
		this.HovE = -1;
		if (hov) {
			for (var i in this.Values) {
				if (x > this.PValues[i] && x < this.PValues[i] + this.LValues[i]) this.HovE = i;
			}
		}
		if (this.hover != hov || this.HovE != HE) {
			if (this.HovE > 0 && this.tooltip) {
				tooltip.Text = this.list[this.HovE];
				tooltip.Activate();
			} else tooltip.Deactivate();
			window.SetCursor(this.HovE >= 0 ? IDC_HAND : IDC_ARROW);
			this.hover = hov;
			ommRepaint = true;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			if (this.HovE >= 0) {
				this.ActE = this.HovE;
				this.Values[this.HovE] = !this.Values[this.HovE];
				this.Save();
				omluRepaint = true;
			}
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover) {
			var n = -1;
			for (var i in this.list) {
				if (this.PValues[i] < x && x < (this.PValues[i] + this.LValues[i])) n = i;
			}
			if (this.rbfunc && n >= 0) {
				this.rbfunc(x, y, n, this.PValues[n], this.y, this.LValues[n], this.THeight);
				FormsAndButtons.Reload();
				return true;
			}
		}
	}
	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		this.ActE = 0;
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_RETURN:
				this.Values[this.ActE] = !this.Values[this.ActE]
				this.Save();
				break;
			case VK_END:
				this.ActE = this.Values.length - 1;
				break;
			case VK_HOME:
				this.ActE = 0;
				break;
			case VK_LEFT:
				this.ActE--;
				if (this.ActE < 0) FormsAndButtons.ActivatePrevious();
				break;
			case VK_UP:
				FormsAndButtons.ActivatePrevious();
				break;
			case VK_RIGHT:
				this.ActE++;
				if (this.ActE >= this.Values.length) FormsAndButtons.ActivateNext();
				break;
			case VK_DOWN:
				FormsAndButtons.ActivateNext();
				break;
			}
			okdRepaint = true;
		}
	}

	this.CalcValues = function () {
		for (var i in this.list) {
			this.Values[i] = false;
			for (var j in this.SValues) {
				if (this.list[i] == this.SValues[j]) this.Values[i] = true;
			}
		}
	}

	this.Save = function () {
		this.SValues = [];
		var t = 0;
		for (var i in this.Values) {
			if (this.Values[i]) {
				this.SValues[t] = this.list[i];
				t++;
			}
		}
		if (this.func) {
			this.func(this.SValues);
			FormsAndButtons.Reload();
		}
	}
}

FormKeyWordsHS.prototype = new FormProto;
FormKeyWordsHS.prototype.constructor = FormKeyWordsHS;

function FormKeyWordsHS(x, y, w, skip, font, listfunc, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.mw = Math.max(0, eval(w));
	this.w = null;
	this.skip = skip;
	this.Value = null;
	this.SValue = null;
	this.PValues = [];
	this.LValues = [];
	this.Font = font;
	this.listfunc = listfunc;
	this.list = [];
	this.func = func;
	this.rbfunc = rbfunc;
	this.THeight = null;
	this.HovE = null;
	this.ActE = 0;
	this.tooltip = false;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
		this.mw = Math.max(0, eval(w));
	}
	this.draw = function (gr) {
		if (this.reload) {
			if (this.listfunc) this.list = this.listfunc();
			if (this.func) {
				var str = this.func();
				if (str != this.SValue) {
					this.SValue = str;
					this.CalcValues();
				}
			}
			this.reload = false;
		}
		var tmpx = 0;
		this.PValues = [];
		this.LValues = [];
		for (var i in this.list) {
			if (i > 0) tmpx += marginx;
			this.PValues[i] = this.x + tmpx
			this.LValues[i] = gr.CalcTextWidth(this.list[i], this.Font);
			tmpx += this.LValues[i];
		}
		this.w = tmpx;
		if (this.w > this.mw && this.mw > 0) {
			var d = (this.w - this.mw) / this.list.length;
			this.w = this.mw;
			this.tooltip = true;
		} else {
			var d = 0;
			this.tooltip = false;
		}
		for (var i in this.list) {
			this.PValues[i] = this.PValues[i] - i * d;
			this.LValues[i] = this.LValues[i] - d;
			var colour = this.Value == i ? textcolour : uncheckedcolour;
			gr.GdiDrawText(this.list[i], this.Font, colour, this.PValues[i], this.y, this.LValues[i], this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
			if (this.active && this.ActE == i) gr.DrawRect(this.PValues[i] - marginx / 2, this.y - marginy / 2, this.LValues[i] + marginx, this.THeight + marginy, 1, activecolour);
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.w + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.w) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		var HE = this.HovE;
		this.HovE = -1;
		if (hov) {
			for (var i in this.list) {
				if (x > this.PValues[i] && x < this.PValues[i] + this.LValues[i]) this.HovE = i;
			}
		}
		if (this.hover != hov || this.HovE != HE) {
			if (this.HovE > 0 && this.tooltip) {
				tooltip.Text = this.list[this.HovE];
				tooltip.Activate();
			} else tooltip.Deactivate();
			window.SetCursor(this.HovE >= 0 ? IDC_HAND : IDC_ARROW);
			this.hover = hov;
			ommRepaint = true;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			if (this.HovE >= 0) {
				this.ActE = this.HovE;
				if (this.Value == this.HovE) {
					this.SValue = "";
					if (this.func) {
						this.func("");
						FormsAndButtons.Reload();
					}
					this.Value = -1;
				} else {
					this.SValue = this.list[this.HovE];
					if (this.func) {
						this.func(this.SValue);
						FormsAndButtons.Reload();
					}
					this.Value = this.HovE;
				}
				omluRepaint = true;
			}
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover) {
			var n = -1;
			for (var i in this.list) {
				if (this.PValues[i] < x && x < (this.PValues[i] + this.LValues[i])) n = i;
			}
			if (this.rbfunc && n >= 0) {
				this.rbfunc(x, y, n, this.PValues[n], this.y, this.LValues[n], this.THeight);
				FormsAndButtons.Reload();
				return true;
			}
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		this.ActE = 0;
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_RETURN:
				if (this.Value == this.ActE) {
					this.SValue = "";
					if (this.func) {
						this.func("");
						FormsAndButtons.Reload();
					}
					this.Value = -1;
				} else {
					this.SValue = this.list[this.ActE];
					if (this.func) {
						this.func(this.SValue);
						FormsAndButtons.Reload();
					}
					this.Value = this.ActE;
				}
				break;
			case VK_END:
				this.ActE = this.list.length - 1;
				break;
			case VK_HOME:
				this.ActE = 0;
				break;
			case VK_LEFT:
				this.ActE--;
				if (this.ActE < 0) FormsAndButtons.ActivatePrevious();
				break;
			case VK_UP:
				FormsAndButtons.ActivatePrevious();
				break;
			case VK_RIGHT:
				this.ActE++;
				if (this.ActE >= this.list.length) FormsAndButtons.ActivateNext();
				break;
			case VK_DOWN:
				FormsAndButtons.ActivateNext();
				break;
			}
			okdRepaint = true;
		}
	}

	this.CalcValues = function () {
		this.Value = -1;
		for (var i in this.list) {
			if (this.list[i] == this.SValue) this.Value = i;
		}
	}
}

FormRating.prototype = new FormProto;
FormRating.prototype.constructor = FormRating;

function FormRating(x, y, skip, font, max, func, rbfunc) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.w = null;
	this.skip = skip;
	this.Value = null;
	this.SValue = null;
	this.Font = font;
	this.max = max;
	this.func = func;
	this.rbfunc = rbfunc;
	this.THeight = null;
	this.HovE = null;
	this.ActE = 0;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
	}

	this.draw = function (gr) {
		if (this.reload) {
			if (this.func) {
				var val = this.func();
				if (val != this.SValue) {
					this.Value = parseInt(val);
					this.SValue = val;
					this.ActE = (isNaN(parseInt(this.Value)) || this.Value == 0) ? 0 : parseInt(this.Value) - 1;
				}
			}
			this.reload = false;
		}
		for (var i = 0; i < this.max; i++) {
			if (i < this.Value) {
				var str = ratingon;
				colour = textcolour;
			} else {
				var str = ratingoff;
				colour = uncheckedcolour;
			}
			if (this.active && this.ActE == i) colour = activecolour;
			gr.GdiDrawText(str, this.Font, colour, this.x + i * this.THeight, this.y, this.THeight, this.THeight, DT_CENTER | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT | DT_END_ELLIPSIS);
		}
		if (this.active) gr.DrawRect(this.x - marginx, this.y - marginy, this.max * this.THeight + 2 * marginx, this.THeight + 2 * marginy, 1, activecolour);
	}

	this.on_mousemove = function (x, y) {
		var hov = (this.x < x && x < (this.x + this.max * this.THeight) && this.y < y && y < (this.y + this.THeight)) ? true : false;
		var HE = this.HovE;
		this.HovE = 0;
		if (hov) {
			for (var i = 0; i < this.max; i++) {
				if (x > this.x + i * this.THeight && x < this.x + (i + 1) * this.THeight) this.HovE = i + 1;
			}
		}
		if (this.hover != hov || this.HovE != HE) {
			window.SetCursor(this.HovE > 0 ? IDC_HAND : IDC_ARROW);
			this.hover = hov;
			ommRepaint = true;
		}
	}

	this.on_mouse_lbtn_up = function (x, y) {
		if (this.hover) {
			if (!this.skip) this.active = true;
			if (this.HovE >= 0) {
				this.ActE = this.HovE;
				if (this.Value == this.HovE) {
					this.Value = 0;
					this.SValue = 0;
					if (this.func) {
						this.func(0);
						FormsAndButtons.Reload();
					}
				} else {
					this.Value = this.HovE;
					this.SValue = this.HovE;
					if (this.func) {
						this.func(this.Value);
						FormsAndButtons.Reload();
					}
				}
				omluRepaint = true;
			}
		}
		return this.hover;
	}

	this.on_mouse_rbtn_up = function (x, y) {
		if (this.hover) {
			var n = -1;
			for (var i = 0; i < this.max; i++) {
				if ((this.x + i * this.THeight) < x && x < (this.x + (i + 1) * this.THeight)) n = i;
			}
			if (this.rbfunc && n >= 0) {
				this.rbfunc(x, y, n, this.x + i * this.THeight, this.y, this.THeight, this.THeight);
				FormsAndButtons.Reload();
				return true;
			}
		}
		if (this.hover && this.rbfunc) {
			this.rbfunc(x, y);
			FormsAndButtons.Reload();
			return true;
		}
	}

	this.on_activate = function (act) {
		this.active = this.skip ? false : act;
		this.ActE = (isNaN(parseInt(this.Value)) || this.Value == 0) ? 0 : parseInt(this.Value) - 1;
	}

	this.on_key = function (vkey, mask) {
		if (mask == KMask.none) {
			switch (vkey) {
			case VK_RETURN:
				if (this.Value == this.ActE + 1) {
					this.Value = 0;
					this.SValue = 0;
					if (this.func) {
						this.func(0);
						FormsAndButtons.Reload();
					}
				} else {
					this.Value = this.ActE + 1;
					this.SValue = this.ActE + 1;
					if (this.func) {
						this.func(this.Value);
						FormsAndButtons.Reload();
					}
				}
				break;
			case VK_END:
				this.ActE = this.max - 1;
				break;
			case VK_HOME:
				this.ActE = 0;
				break;
			case VK_LEFT:
				this.ActE--;
				if (this.ActE < 0) FormsAndButtons.ActivatePrevious();
				break;
			case VK_UP:
				FormsAndButtons.ActivatePrevious();
				break;
			case VK_RIGHT:
				this.ActE++;
				if (this.ActE >= this.max) FormsAndButtons.ActivateNext();
				break;
			case VK_DOWN:
				FormsAndButtons.ActivateNext();
				break;
			}
			okdRepaint = true;
		}
	}
}

FormString.prototype = new FormProto;
FormString.prototype.constructor = FormString;

function FormString(x, y, font, colour, str) {
	this.x = Math.max(0, eval(x));
	this.y = Math.max(0, eval(y));
	this.Font = font;
	this.colour = colour;
	this.Text = str;
	this.TWidth = null;
	this.THeight = null;
	var arr = temp_gr.GdiDrawText(".", this.Font, 0, 0, 0, 9999999, 9999999, DT_VCENTER | DT_CALCRECT).toArray();
	this.THeight = arr[3] - arr[1];
	this.TWidth = temp_gr.CalcTextWidth(this.Text, this.Font);

	this.resize = function () {
		this.x = Math.max(0, eval(x));
		this.y = Math.max(0, eval(y));
	}

	this.draw = function (gr) {
		gr.GdiDrawText(this.Text, this.Font, this.colour, this.x, this.y, this.TWidth, this.THeight, DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_CALCRECT);
	}
}

//--------------------
// HELPERS
//--------------------

VK_BACK = 0x08;
VK_RETURN = 0x0D;
VK_SHIFT = 0x10;
VK_CONTROL = 0x11;
VK_ALT = 0x12;
VK_ESCAPE = 0x1B;
VK_PGUP = 0x21;
VK_PGDN = 0x22;
VK_END = 0x23;
VK_HOME = 0x24;
VK_LEFT = 0x25;
VK_UP = 0x26;
VK_RIGHT = 0x27;
VK_DOWN = 0x28;
VK_INSERT = 0x2D;
VK_DELETE = 0x2E;

KMask = {
    none: 0,
    ctrl: 1,
    shift: 2,
    ctrlshift: 3,
    ctrlalt: 4,
    ctrlaltshift: 5
}

function GetKeyboardMask() {
    var c = utils.IsKeyPressed(VK_CONTROL) ? true : false;
    var a = utils.IsKeyPressed(VK_ALT) ? true : false;
    var s = utils.IsKeyPressed(VK_SHIFT) ? true : false;
    var ret = KMask.none;
    if (c && !a && !s) ret = KMask.ctrl;
    if (!c && !a && s) ret = KMask.shift;
    if (c && !a && s) ret = KMask.ctrlshift;
    if (c && a && !s) ret = KMask.ctrlalt;
    if (c && a && s) ret = KMask.ctrlaltshift;
    return ret
}

//--------------------

var ww = 0
var wh = 0;
var px = 0;
var py = 0;
var pw = 0;
var ph = 0;

var ommRepaint = false;
var omldRepaint = false;
var omluRepaint = false;
var omrdRepaint = false;
var omruRepaint = false;
var omwRepaint = false;
var okdRepaint = false;
var ocRepaint = false;
var ofRepaint = false;

var temp_bmp = gdi.CreateImage(1, 1);
var temp_gr = temp_bmp.GetGraphics();
var tooltip = window.CreateTooltip();
var ctimer = window.CreateTimerInterval(1000);
window.DlgCode = DLGC_WANTALLKEYS | DLGC_WANTCHARS;

//--------------------
// CALLBACKS
//--------------------

function forms_on_size() {
	ww = window.Width;
	wh = window.Height;
	FormsAndButtons.SetPopUp();
	for (var i in FormsAndButtons.list) {
		FormsAndButtons.list[i].resize();
	}
	for (var i in FormsAndButtons.popuplist) {
		FormsAndButtons.popuplist[i].resize();
	}
	SimpleInput.resize();
	FormsAndButtons.resize = true;
}

function forms_on_paint(gr) {
	gr.SetInterpolationMode(7);
	gr.SetSmoothingMode(4);
	gr.FillSolidRect(-1, -1, ww + 1, wh + 1, backcolour);
	for (var i in FormsAndButtons.list) {
		FormsAndButtons.list[i].draw(gr);
	}
	if (FormsAndButtons.popup) {
		gr.FillSolidRect(-1, -1, ww + 1, wh + 1, setAlpha(backcolour, 192));
		gr.FillSolidRect(px - 1, py - 1, pw + 1, ph + 1, backcolour);
		gr.DrawRect(px + 10, py + 10, pw - 20, ph - 20, 1, textcolour);
		for (var i in FormsAndButtons.popuplist) {
			FormsAndButtons.popuplist[i].draw(gr);
		}
	}
	SimpleInput.draw(gr);
}

function forms_on_mouse_move(x, y) {
	ommRepaint = false;
	SimpleInput.on_mousemove(x, y);
	if (!SimpleInput.hover) {
		if (FormsAndButtons.popup) {
			for (var i in FormsAndButtons.list) {
				if (i >= FormsAndButtons.list.length) break;
				FormsAndButtons.list[i].on_mousemove(-1, -1);
			}
			for (var i in FormsAndButtons.popuplist) {
				if (i >= FormsAndButtons.popuplist.length) break;
				FormsAndButtons.popuplist[i].on_mousemove(x, y);
			}
		} else {
			for (var i in FormsAndButtons.list) {
				if (i >= FormsAndButtons.list.length) break;
				FormsAndButtons.list[i].on_mousemove(x, y);
			}
			for (var i in FormsAndButtons.popuplist) {
				if (i >= FormsAndButtons.popuplist.length) break;
				FormsAndButtons.popuplist[i].on_mousemove(-1, -1);
			}
		}
	}
	if (ommRepaint) window.Repaint();
}

function forms_on_mouse_lbtn_down(x, y) {
	omldRepaint = false;
	SimpleInput.lbtnd = SimpleInput.hover;
	if (SimpleInput.hover) SimpleInput.on_mouse_lbtn_down(x, y);
	for (var i in FormsAndButtons.list) {
		if (i >= FormsAndButtons.list.length) break;
		FormsAndButtons.list[i].lbtnd = FormsAndButtons.list[i].hover;
		if (FormsAndButtons.list[i].hover) FormsAndButtons.list[i].on_mouse_lbtn_down(x, y);
	}
	for (var i in FormsAndButtons.popuplist) {
		if (i >= FormsAndButtons.popuplist.length) break;
		FormsAndButtons.popuplist[i].lbtnd = FormsAndButtons.popuplist[i].hover;
		if (FormsAndButtons.popuplist[i].hover) FormsAndButtons.popuplist[i].on_mouse_lbtn_down(x, y);
	}
	if (omldRepaint) window.Repaint();
}

function forms_on_mouse_lbtn_up(x, y) {
	omluRepaint = false;
	if (SimpleInput.lbtnd) {
		SimpleInput.lbtnd = false;
		SimpleInput.on_mouse_lbtn_up(x, y);
	}
	for (var i in FormsAndButtons.list) {
		if (i >= FormsAndButtons.list.length) break;
		if (FormsAndButtons.list[i].lbtnd) {
			FormsAndButtons.list[i].lbtnd = false;
			if (FormsAndButtons.list[i].on_mouse_lbtn_up(x, y)) FormsAndButtons.activeobj = i;
		} else FormsAndButtons.list[i].on_activate(false);
	}
	for (var i in FormsAndButtons.popuplist) {
		if (i >= FormsAndButtons.popuplist.length) break;
		if (FormsAndButtons.popuplist[i].lbtnd) {
			FormsAndButtons.popuplist[i].lbtnd = false;
			if (FormsAndButtons.popuplist[i].on_mouse_lbtn_up(x, y)) FormsAndButtons.activeobj = i;
		} else FormsAndButtons.popuplist[i].on_activate(false);
	}
	if (omluRepaint) window.Repaint();
}

function forms_on_mouse_rbtn_down(x, y) {
	omrdRepaint = false;
	SimpleInput.rbtnd = SimpleInput.hover;
	if (SimpleInput.hover) SimpleInput.on_mouse_rbtn_down(x, y);
	for (var i in FormsAndButtons.list) {
		if (i >= FormsAndButtons.list.length) break;
		FormsAndButtons.list[i].rbtnd = FormsAndButtons.list[i].hover;
		if (FormsAndButtons.list[i].hover) FormsAndButtons.list[i].on_mouse_rbtn_down(x, y);
	}
	for (var i in FormsAndButtons.popuplist) {
		if (i >= FormsAndButtons.popuplist.length) break;
		FormsAndButtons.popuplist[i].rbtnd = FormsAndButtons.popuplist[i].hover;
		if (FormsAndButtons.popuplist[i].hover) FormsAndButtons.popuplist[i].on_mouse_rbtn_down(x, y);
	}
	if (omrdRepaint) window.Repaint();
}

function forms_on_mouse_rbtn_up(x, y) {
	omruRepaint = false;
	var tmp = false;
	if (SimpleInput.rbtnd) {
		SimpleInput.rbtnd = false;
		if (SimpleInput.on_mouse_rbtn_up(x, y)) tmp = true;
	}
	for (var i in FormsAndButtons.list) {
		if (i >= FormsAndButtons.list.length) break;
		if (FormsAndButtons.list[i].rbtnd) {
			FormsAndButtons.list[i].rbtnd = false;
			if (FormsAndButtons.list[i].on_mouse_rbtn_up(x, y)) {tmp = true; FormsAndButtons.activeobj = i};
		} else FormsAndButtons.list[i].on_activate(false);
	}
	for (var i in FormsAndButtons.popuplist) {
		if (i >= FormsAndButtons.popuplist.length) break;
		if (FormsAndButtons.popuplist[i].rbtnd) {
			FormsAndButtons.popuplist[i].rbtnd = false;
			if (FormsAndButtons.popuplist[i].on_mouse_rbtn_up(x, y)) {tmp = true; FormsAndButtons.activeobj = i};
		} else FormsAndButtons.popuplist[i].on_activate(false);
	}
	if (omruRepaint) window.Repaint();
	return tmp;
}

function forms_on_mouse_wheel(step) {
	omwRepaint = false;
	var objects = FormsAndButtons.popup ? FormsAndButtons.popuplist : FormsAndButtons.list;
	for (var i in objects) {
		if (i >= objects.length) break;
		objects[i].on_mouse_wheel(step);
	}
	if (omwRepaint) window.Repaint();
}

function forms_on_key_down(vkey) {
	okdRepaint = false;
	if (SimpleInput.active) SimpleInput.on_key(vkey, GetKeyboardMask());
	else {
		var objects = FormsAndButtons.popup ? FormsAndButtons.popuplist : FormsAndButtons.list;
		if (FormsAndButtons.activeobj) objects[FormsAndButtons.activeobj].on_key(vkey, GetKeyboardMask());
	}
	if (okdRepaint) window.Repaint();
}

function forms_on_char(code) {
	ocRepaint = false;
	if (SimpleInput.active) SimpleInput.on_char(code, GetKeyboardMask());
	else {
		var objects = FormsAndButtons.popup ? FormsAndButtons.popuplist : FormsAndButtons.list;
		if (FormsAndButtons.activeobj) objects[FormsAndButtons.activeobj].on_char(code, GetKeyboardMask());
	}
	if (ocRepaint) window.Repaint();
}

function forms_on_focus(is_focused) {
	ofRepaint = false;
	if (is_focused) {
		FormsAndButtons.ActivateNext();
	} else {
		SimpleInput.Close();
		for (var i in FormsAndButtons.list) {
			FormsAndButtons.list[i].on_activate(false);
		}
		for (var i in FormsAndButtons.popuplist) {
			FormsAndButtons.popuplist[i].on_activate(false);
		}
	}
	window.Repaint();
}

function forms_on_timer(id) {
    if (id == ctimer.ID) {
        if (FormsAndButtons.resize) {
            FormsAndButtons.resize = false;
            window.Repaint();
        }
        FormsAndButtons.cvisible = !FormsAndButtons.cvisible;
        window.RepaintRect(FormsAndButtons.crpx, FormsAndButtons.crpy, FormsAndButtons.crpw, FormsAndButtons.crph);
    }
}
